import numpy as np
import numpy.ma as ma
import os
from itertools import islice
from particle import literals as lp
from DarkNews import logger
try:
    import importlib.resources as resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as resources


from DarkNews.const_dics import fourier_bessel_dic
from DarkNews import const


class NuclearTarget:
    def __init__(self, name):
        """
        Main DarkNews class for the nucleus to be used in a neutrino scattering event.

        It contains target properties like number of protons, neutrons, informations on the mass, as well 
        as all the Nuclear Data Table information on:
            nuclear_Eb,
            atomic_Eb,
            atomic_mass,
            excess_mass,
            nuclear_mass,
            beta_decay_energy.

        It also provides the nuclear form factors as functions to

        Its daughter class, BoundNucleon(), handles the nucleons inside the nucleus.

        Args:
            name (str): name of the target. Can be either "electron" or the name of the element
            following the Nuclear Data Table format (e.g. C12, Pb208).

        """

        self.name = name

        #####################################
        # free electron
        if name == "electron":
            self.is_hadron = False
            self.is_nucleus = False
            self.is_proton = False
            self.is_neutron = False
            self.is_free_nucleon = False

            self.mass = const.m_e
            self.charge = 1
            self.Z = 0
            self.N = 0
            self.A = 0
            self.pdgid = 11
        #####################################
        # hadronic *nuclear* target
        else:

            # Using the global dictionary of elements defined in nuclear_tools
            # Set all items as attributes of the class
            for k, v in elements_dic[name].items():
                setattr(self, k, v)
            self.mass = self.nuclear_mass
            self.charge = self.Z

            self.is_hadron = True
            self.is_nucleus = self.A > 1
            self.is_proton = self.Z == 1 and self.A == 1
            self.is_neutron = self.N == 1 and self.A == 1

            self.is_nucleon = self.is_neutron or self.is_proton
            self.is_free_nucleon = self.A == 1
            self.is_bound_nucleon = False

            if self.is_nucleus:
                # no hyperons and always ground state
                self.pdgid = int(f"100{self.Z:03d}{self.A:03d}0")
            elif self.is_neutron:
                self.pdgid = lp.neutron.pdgid
            elif self.is_proton:
                self.pdgid = lp.proton.pdgid
            else:
                logger.error(f"Error. Could not find the PDG ID of {self.name}.")
                raise ValueError

            if self.is_neutron and self.is_free_nucleon:
                logger.error(f"Error. Target {self.name} is a free neutron.")
                raise ValueError

            self.tau3 = self.Z * 2 - 1  # isospin +1 proton / -1 neutron

            assign_form_factors(self)

    # hadronic *constituent* target
    def get_constituent_nucleon(self, name):
        return self.BoundNucleon(self, name)

    class BoundNucleon:
        """ for scattering on bound nucleon in the nuclear target

        Inner Class

        Args:
            nucleus: nucleus of which this particle is bound into (always the outer class)
            name: 'proton' or 'neutron'
        """

        def __init__(self, nucleus, name):

            self.nucleus = nucleus
            self.A = int(name == "proton" or name == "neutron")
            self.Z = int(name == "proton")
            self.N = int(name == "neutron")
            self.charge = self.Z
            self.mass = const.m_proton
            self.name = f"{name}_in_{nucleus.name}"

            self.is_hadron = True
            self.is_nucleus = False
            self.is_neutron = self.N == 1 and self.A == 1
            self.is_proton = self.Z == 1 and self.A == 1

            self.is_nucleon = self.is_neutron or self.is_proton
            self.is_false_nucleon = False
            self.is_bound_nucleon = True

            if self.is_neutron:
                self.pdgid = lp.neutron.pdgid
            elif self.is_proton:
                self.pdgid = lp.proton.pdgid
            else:
                logger.error(f"Error. Could not find the PDG ID of {self.name}.")
                raise ValueError

            self.tau3 = self.Z * 2 - 1  # isospin +1 proton / -1 neutron

            assign_form_factors(self)


def assign_form_factors(target):
    """
        Here we define nuclear form factors following:
        http://discovery.phys.virginia.edu/research/groups/ncd/index.html

        When available, we use data from Nuclear Data Tables (74, 87, and 95), stored in "include/aux_data/mass20.txt":
            "The Ame2020 atomic mass evaluation (I)"   by W.J.Huang, M.Wang, F.G.Kondev, G.Audi and S.Naimi
                Chinese Physics C45, 030002, March 2021.
            "The Ame2020 atomic mass evaluation (II)"  by M.Wang, W.J.Huang, F.G.Kondev, G.Audi and S.Naimi
                Chinese Physics C45, 030003, March 2021.

        Element properties are stored in elements_dic.To access individual elements we use the format:

            key = 'name+A', e.g. key = 'Pb208' or 'C12'.

        All units in GeV, except otherwise specified.

        Args:
            target (DarkNews.nuclear_tools.Target): instance of main DarkNews target object.
    """

    # Nucleus
    if target.is_nucleus:
        try:
            a = fourier_bessel_dic[target.name.lower()]  ## stored with lower case formatting
            fcoh = lambda x: nuclear_F1_fourier_bessel_EM(x, a)
        except KeyError:
            logger.warning(f"Warning: nuclear density for {target.name} not tabulated in Nuclear Data Table. Using symmetrized Fermi form factor instead.")
            fcoh = lambda x: nuclear_F1_Fsym_EM(x, target.A)
        except:
            logger.warning(f"Warning: could not compute the nuclear form factor for {target.name}. Taking it to be vanishing.")
            fcoh = lambda x: 0

        ### FIX ME -- No nuclear magnetic moments so far
        target.F1_EM = fcoh  # Dirac FF
        target.F2_EM = lambda x: 0.0  # Pauli FF

        ### FIX ME -- need to find the correct NC form factor
        # target.F1_NC = lambda x: nuclear_F1_FHelmz_NC(x, target.A) # Dirac FF
        target.F1_NC = fcoh  # Dirac FF
        target.F2_NC = lambda x: 0.0  # Pauli FF
        target.F3_NC = lambda x: 0.0  # Axial FF

    # Nucleons
    elif target.is_nucleon:

        target.F1_EM = lambda x: nucleon_F1_EM(x, target.tau3)  # Dirac FF
        target.F2_EM = lambda x: nucleon_F2_EM(x, target.tau3)  # Pauli FF

        target.F1_NC = lambda x: nucleon_F1_NC(x, target.tau3)  # Dirac FF
        target.F2_NC = lambda x: nucleon_F2_NC(x, target.tau3)  # Pauli FF
        target.F3_NC = lambda x: nucleon_F3_NC(x, target.tau3)  # Axial FF

    else:
        logger.error(f"Could not find hadronic target {target.name}.")
        exit(0)


#####################################
# Partdon Distribution Functions
def assign_PDFs(target):

    # Nucleus
    if target.is_nucleus:
        try:
            a = fourier_bessel_dic[target.name.lower()]  ## stored with lower case formatting
            fcoh = lambda x: nuclear_F1_fourier_bessel_EM(x, a)
        except KeyError:
            logger.warning(f"Warning: nuclear density for {target.name} not tabulated in Nuclear Data Table. Using symmetrized Fermi form factor instead.")
            fcoh = lambda x: nuclear_F1_Fsym_EM(x, target.A)
        except:
            logger.warning(f"Warning: could not compute the nuclear form factor for {target.name}. Taking it to be vanishing.")
            fcoh = lambda x: 0

        ### FIX ME -- No nuclear magnetic moments so far
        target.F1_EM = fcoh  # Dirac FF
        target.F2_EM = lambda x: 0.0  # Pauli FF

        ### FIX ME -- need to find the correct NC form factor
        # target.F1_NC = lambda x: nuclear_F1_FHelmz_NC(x, target.A) # Dirac FF
        target.F1_NC = fcoh  # Dirac FF
        target.F2_NC = lambda x: 0.0  # Pauli FF
        target.F3_NC = lambda x: 0.0  # Axial FF

    # Nucleons
    elif target.is_nucleon:

        target.F1_EM = lambda x: nucleon_F1_EM(x, target.tau3)  # Dirac FF
        target.F2_EM = lambda x: nucleon_F2_EM(x, target.tau3)  # Pauli FF

        target.F1_NC = lambda x: nucleon_F1_NC(x, target.tau3)  # Dirac FF
        target.F2_NC = lambda x: nucleon_F2_NC(x, target.tau3)  # Pauli FF
        target.F3_NC = lambda x: nucleon_F3_NC(x, target.tau3)  # Axial FF

    else:
        logger.error(f"Could not find hadronic target {target.name}.")
        exit(0)


# covering the zero case 0
def sph_bessel_0(x):
    x = np.array(x)
    tolerance = x > 0.0
    j0 = ma.masked_array(data=np.ma.sin(x) / x, mask=~tolerance, fill_value=1.0)
    return j0.filled()


# all units in fm
def fourier_bessel_form_factor_terms(q, R, n):
    x = q * R
    return sph_bessel_0(x) * R ** 3 / ((n * np.pi) ** 2 - x ** 2) * (-1) ** n


# all units in fm
def fourier_bessel_integral_terms(R, n):
    return R ** 3 * (-1) ** n / np.pi ** 2 / n ** 2


# Q2 in GeV^2 and other units in fm
def nuclear_F1_fourier_bessel_EM(Q2, array_coeff):
    q = np.sqrt(Q2) * const.fm_to_GeV
    R = array_coeff[-1]
    ai = np.array(array_coeff[:-1])
    nonzero_terms = np.nonzero(ai)[0]
    # compute expansion for non-zero terms
    expansion = np.zeros((np.size(q)))
    for i in nonzero_terms:
        n = i + 1
        expansion += fourier_bessel_form_factor_terms(q, R, n) * ai[i]
    Q_total = np.sum(fourier_bessel_integral_terms(R, nonzero_terms + 1) * ai[nonzero_terms])
    return np.abs(expansion / Q_total)


#####################################
# Nested dic containing all elements
# approximate formula in D. Lunney, J.M. Pearson and C. Thibault, Rev. Mod. Phys.75, 1021 (2003)
def electron_binding_energy(Z):
    return (14.4381 * Z ** 2.39 + 1.55468e-6 * Z ** 5.35) * 1e-9  # GeV


elements_dic = {}
hydrogen_Eb = 13.5981e-9  # GeV
atomic_unit = 0.9314941024228  # mass of Carbon12 in GeV / 12
with resources.open_text("DarkNews.include.aux_data", "mass20_1.txt") as ame:
    # Read lines in file starting at line 36
    for line in islice(ame, 36, None):

        Z = int(line[12:15])

        if Z < 93:  ## no support for heavier elements due to the
            name = "{}{}".format(line[20:22].strip(), int(line[16:19]))
            elements_dic[name] = {}

            elements_dic[name]["name"] = name
            elements_dic[name]["Z"] = Z
            elements_dic[name]["N"] = int(line[6:10])
            elements_dic[name]["A"] = int(line[16:20])

            # elements_dic[name]['atomic_Eb'] = float(line[56:61] + '.' + line[62:67])*1e-6*elements_dic[name]['A']
            elements_dic[name]["atomic_Eb"] = electron_binding_energy(Z)

            elements_dic[name]["nuclear_Eb"] = float(line[56:61] + "." + line[62:67]) * 1e-6 * elements_dic[name]["A"]
            # elements_dic[name]['atomic_Eb'] + Z*hydrogen_Eb - elements_dic[name]['electronic_Eb']

            # micro-mu = 1e-6 mu in GeV*u
            elements_dic[name]["atomic_mass"] = (float(line[106:109]) + 1e-6 * float(line[110:116] + "." + line[117:124])) * atomic_unit
            elements_dic[name]["excess_mass"] = float(line[30:35] + "." + line[36:42]) * 1e-6

            # nuclear mass corrected for electron mass and electron binding energy (note that Eb_e << MeV)
            elements_dic[name]["nuclear_mass"] = elements_dic[name]["atomic_mass"] - Z * const.m_e + elements_dic[name]["atomic_Eb"]

            decay = line[82:88] + "." + line[89:94]
            if "*" in decay:
                decay = np.nan
            elements_dic[name]["beta_decay_energy"] = float(decay) * 1e-6


################################################
# NUCLEAR AND NUCLEON FORM FACTORS
MAG_N = -1.913
MAG_P = 2.792
## dipole parametrization
def D(Q2):
    MV = 0.843  # GeV
    return 1.0 / ((1 + Q2 / MV ** 2) ** 2)


## nucleon
def nucleon_F1_EM(Q2, tau3):
    # pick nucleon mag moment
    MAG = (MAG_P + MAG_N + tau3 * (MAG_P - MAG_N)) / 2.0
    tau = -Q2 / 4.0 / const.m_proton ** 2
    return (D(Q2) - tau * MAG * D(Q2)) / (1 - tau)


def nucleon_F2_EM(Q2, tau3):
    # pick nucleon mag moment
    MAG = (MAG_P + MAG_N + tau3 * (MAG_P - MAG_N)) / 2.0
    tau = -Q2 / 4.0 / const.m_proton ** 2
    return (MAG * D(Q2) - D(Q2)) / (1 - tau)


def nucleon_F1_NC(Q2, tau3):
    tau = -Q2 / 4.0 / const.m_proton ** 2
    f = (0.5 - const.s2w) * (tau3) * (1 - tau * (1 + MAG_P - MAG_N)) / (1 - tau) - const.s2w * (1 - tau * (1 + MAG_P + MAG_N)) / (1 - tau)
    return f * D(Q2)


def nucleon_F2_NC(Q2, tau3):
    tau = -Q2 / 4.0 / const.m_proton ** 2
    f = (0.5 - const.s2w) * (tau3) * (MAG_P - MAG_N) / (1 - tau) - const.s2w * (MAG_P + MAG_N) / (1 - tau)
    return f * D(Q2)


def nucleon_F3_NC(Q2, tau3):
    MA = 1.02  # GeV
    gA = 1.26
    f = gA * tau3 / 2.0 / (1 + Q2 / MA ** 2) ** 2
    return f


## symmetrized fermi nuclear
def nuclear_F1_Fsym_EM(Q2, A):
    Q = np.sqrt(Q2)
    a = 0.523 * const.fm_to_GeV  # GeV^-1
    r0 = 1.03 * (A ** (1.0 / 3.0)) * const.fm_to_GeV  # GeV^-1
    tolerance = Q > 0.0
    clean_FF = ma.masked_array(
        data=3.0 * np.pi * a / (r0 ** 2 + np.pi ** 2 * a ** 2) * 
            (np.pi * a * (1.0 / np.tanh(np.pi * a * Q)) * np.sin(Q * r0) - r0 * np.cos(Q * r0)) /
            (Q * r0 * np.sinh(np.pi * Q * a)),
        mask=~tolerance,
        fill_value=1.0,
    )
    return clean_FF.filled()


def j1(z):
    z = np.array(z)
    tolerance = z > 0.0
    clean_j1 = ma.masked_array(data=np.sin(z) / z / z - np.cos(z) / z, mask=~tolerance, fill_value=0.0)
    return clean_j1.filled()


def nuclear_F1_FHelmz_NC(Q2, A):
    Q = np.sqrt(Q2)
    a = 0.523 * const.fm_to_GeV  # GeV^-1
    s = 0.9 * const.fm_to_GeV  # fm to GeV^-1
    R = 3.9 * const.fm_to_GeV * (A / 40.0) ** (1.0 / 3.0)  # fm to GeV^-1
    return (3 * np.abs(j1(Q * R) / Q / R)) * np.exp(-Q * Q * s * s / 2)



__version__ = "0.1.0"

import sys
from pathlib import Path

"""
    Initializing loggers
"""
import logging

# for debug and error handling
logger = logging.getLogger(__name__ + ".logger")

# for pretty printing
prettyprinter = logging.getLogger(__name__ + ".pretty_printer")
prettyprinter.setLevel(logging.INFO)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setLevel(logging.INFO)
if not prettyprinter.hasHandlers():
    prettyprinter.addHandler(handler)
logger.propagate = False
prettyprinter.propagate = False


"""
    These definition modules make import of main DarkNews classes easier.
    Essentially, it allows DarkNews.XXXX instead of DarkNews.YYYY.XXXX
"""
# Definition modules
from DarkNews.GenLauncher import GenLauncher

from DarkNews import pdg
from DarkNews import const
from DarkNews import fourvec
from DarkNews import phase_space
from DarkNews import parsing_tools

# Experimental setups
from DarkNews import detector
from DarkNews import nuclear_tools

# Physics modules
from DarkNews import decay_rates
from DarkNews import processes
from DarkNews import model

from DarkNews.processes import UpscatteringProcess
from DarkNews.processes import FermionDileptonDecay
from DarkNews.processes import FermionSinglePhotonDecay

# Monte Carlo modules
from DarkNews import MC

# for output of MC
from DarkNews import printer
from DarkNews import geom

from DarkNews import plot_tools



import os
import sys
import pandas as pd
import numpy as np
import dill

from pathlib import Path
from particle import literals as lp

from DarkNews import logger, prettyprinter
from DarkNews import const
from DarkNews import pdg
from DarkNews import Cfourvec as Cfv

import pyarrow.parquet as pq
import pyarrow as pa

import pyhepmc_ng as hep


def print_in_order(x):
    return " ".join(f"{t:.8E}" for t in list(x))


class Printer:
    def __init__(self, df_gen, data_path=None, sparse=False, print_to_float32=False, decay_product="e+e-"):
        """
		Main printer of DarkNews. Can print events contained in the pandas dataframe to files of various types.

		Args:
			df_gen (pd.DataFrame): dataframe with the generated events.

			data_path (str, optional): path to be used to save the event files. Defaults to the "data_path" attribute of df_gen

			sparse (bool, optional): if True, save only the neutrino energy, charged lepton or photon momenta, and weights to save storage space.
									Not supported for HEPevt.
									Defaults to False.

			print_to_float32 (bool, optional): If true downgrade floats to float32 to save storage space. Only relevant when sparse is True.
												Defaults to False.

			decay_product (str, optional): what decay products are being used. Choices = ["e+e-", "mu+mu-", "photon"]. Defaults to 'e+e-'.

		"""

        # main DataFrame
        self.df_gen = df_gen
        self.decay_product = decay_product

        if data_path:
            self.data_path = data_path
        else:
            self.data_path = self.df_gen.attrs["data_path"]

        # file name and path (without extension)
        self.out_file_name = self.create_dir()

        self._sparse = sparse
        if self._sparse:
            self.sparse = self._sparse

        self._print_to_float32 = print_to_float32
        if self._print_to_float32:
            self.print_to_float32 = self._print_to_float32

        self._kinematics_in_np_arrays = False

    @property
    def sparse(self):
        return self._sparse

    @sparse.setter
    def sparse(self, new_value):
        self._sparse = new_value
        if self._sparse:
            self.df_sparse = self.get_sparse_df(self.df_gen)

    @property
    def print_to_float32(self):
        return self._print_to_float32

    @print_to_float32.setter
    def print_to_float32(self, new_value):
        self._print_to_float32 = new_value
        if self._print_to_float32:
            if self.sparse:
                self.df_sparse = self.df_sparse.apply(pd.to_numeric, downcast="float")
            else:
                raise ValueError("Can only downgrade dataframe to float32 in sparse mode.")

    # Create target directory if it doesn't exist
    def create_dir(self):
        if not os.path.exists(self.data_path):
            os.makedirs(self.data_path)
        return Path(self.data_path)

    # Keep only Enu, charged leptons, photons, and weights
    def get_sparse_df(self, df_gen):
        # keep neutrino energy
        keep_cols = ["P_projectile"]
        for col in df_gen.columns.levels[0]:
            if "_ell_" in col or "_photon" in col or "w_decay" in col or "w_event" in col or "N_parent" in col:
                keep_cols.append(col)

        # remove unused levels of MultiIndex
        df = df_gen[keep_cols].drop([("P_projectile", "1"), ("P_projectile", "2"), ("P_projectile", "3")], axis=1)
        df.columns = df.columns.remove_unused_levels()
        return df

    def print_events_to_ndarray(self, **kwargs):
        """ 
			Print to numpy array file (.npy) 
		
		"""
        kwargs["allow_pickle"] = kwargs.get("allow_pickle", False)

        if self.sparse:
            if self.print_to_float32:
                self.array_gen = self.df_sparse.to_numpy(dtype=np.float32)
            else:
                self.array_gen = self.df_sparse.to_numpy(dtype=np.float64)
            cols = [f"{v[0]}_{v[1]}" if v[1] else f"{v[0]}" for v in self.df_sparse.columns.values]
        else:
            # convert to numeric values
            # self.df_gen.loc[self.df_gen['helicity']=='conserving', ['helicity']] = '+1'
            # self.df_gen.loc[self.df_gen['helicity']=='flipping', ['helicity']] = '-1'
            self.df_gen = self.df_gen.replace(to_replace="conserving", value="+1")
            self.df_gen = self.df_gen.replace(to_replace="flipping", value="-1")
            # remove non-numeric entries
            self.df_for_numpy = self.df_gen.drop(["underlying_process", "target", "scattering_regime"], axis=1, level=0)
            cols = [f"{v[0]}_{v[1]}" if v[1] else f"{v[0]}" for v in self.df_for_numpy.columns.values]
            self.array_gen = self.df_for_numpy.to_numpy(dtype=np.float64)

        filename = Path(f"{self.out_file_name}/ndarray.npy").__str__()
        np.save(filename, self.array_gen, **kwargs)
        prettyprinter.info(f"Events in numpy array saved to file successfully:\n{filename}")
        return self.array_gen

    def print_events_to_parquet(self, **kwargs):
        """ 
			Print to pandas DataFrame to parquet file using pyarrow (.parquet) 

			This format cannot save df.attrs to file.

		"""
        # kwargs['engine']=kwargs.get('engine','pyarrow')
        filename = Path(f"{self.out_file_name}/pandas_df.parquet").__str__()
        if self.sparse:
            pq.write_table(pa.Table.from_pandas(self.df_sparse), filename, **kwargs)
            prettyprinter.info(f"Events in sparse pandas dataframe saved to parquet file successfully:\n{filename}")
            return self.df_sparse
        else:
            pq.write_table(pa.Table.from_pandas(self.df_gen), filename, **kwargs)
            prettyprinter.info(f"Events in pandas dataframe saved to parquet file successfully:\n{filename}")
            return self.df_gen

    def print_events_to_pandas(self, **kwargs):
        """ 
			Print to pandas DataFrame pickle file (.pckl)

			This is the only format that allows to save df.attrs to file.
			Using Dill to serialize the Model, Detector, and NuclearTarget classes to file.

		"""
        filename = Path(f"{self.out_file_name}/pandas_df.pckl").__str__()
        if self.sparse:
            dill.dump(self.df_sparse, open(filename, "wb"), **kwargs)
            prettyprinter.info(f"Events in sparse pandas dataframe saved to file successfully:\n{filename}")
            return self.df_sparse
        else:
            # pickles DarkNews classes with support for lambda functions
            dill.dump(self.df_gen, open(filename, "wb"), **kwargs)
            prettyprinter.info(f"Events in pandas dataframe saved to file successfully:\n{filename}")
            return self.df_gen

    def get_unweighted_events(self, nevents, prob_col="w_event_rate", **kwargs):
        """
			Unweigh events in dataframe down to "nevents" using accept-reject method with the weights in "prob_col" of dataframe.
				
			Fails if nevents is smaller than the total number of unweighted events.

			kwargs passed to numpy's random.choice.
		"""
        logger.info(f"Unweighing events down to {nevents} entries.")
        prob = self.df_gen[prob_col] / np.sum(self.df_gen[prob_col])
        if (prob < 0).any():
            logger.error(f"ERROR! Probabily for unweighting contains negative values! Bad weights? {prob_col} < 0.")
        if (prob == 0).any():
            logger.warning(f"WARNING! Discarding zero-valued weights for unweighting. Total of {sum(prob == 0)} of {len(prob)} zero entries for {prob_col}.")
            AccEntries = np.random.choice(self.df_gen.index[prob > 0], size=nevents, p=prob[prob > 0], *kwargs)
        else:
            AccEntries = np.random.choice(self.df_gen.index, size=nevents, p=prob, *kwargs)

        # extract selected entries
        return self.df_gen.filter(AccEntries, axis=0).reset_index()

    def _prepare_kinematics(self, hep_unweight=False, unweighted_hep_events=None):
        """ pre compute the numpy arrays from dataframe to speed up hepmc priting routines. 

		Args:
			hep_unweight (bool, optional): _description_. Defaults to False.
			unweighted_hep_events (_type_, optional): _description_. Defaults to None.
		"""

        if not self._kinematics_in_np_arrays:
            # Unweigh events down to unweighted_hep_events?
            if hep_unweight:
                df_gen = self.get_unweighted_events(nevents=unweighted_hep_events)
            else:
                df_gen = self.df_gen

            # sample size (# of events)
            self.tot_events_to_print = min(len(self.df_gen.index), unweighted_hep_events)

            if not "pos_scatt" in self.df_gen.columns:

                self.df_gen["pos_scatt", "0"] = np.zeros((self.tot_events_to_print, 0))
                self.df_gen["pos_scatt", "1"] = np.zeros((self.tot_events_to_print, 0))
                self.df_gen["pos_scatt", "2"] = np.zeros((self.tot_events_to_print, 0))
                self.df_gen["pos_scatt", "3"] = np.zeros((self.tot_events_to_print, 0))

            if not "pos_decay" in self.df_gen.columns:

                self.df_gen["pos_decay", "0"] = self.df_gen["pos_scatt", "0"]
                self.df_gen["pos_decay", "1"] = self.df_gen["pos_scatt", "1"]
                self.df_gen["pos_decay", "2"] = self.df_gen["pos_scatt", "2"]
                self.df_gen["pos_decay", "3"] = self.df_gen["pos_scatt", "3"]

            # pre-computing some variables

            self.mass_projectile = Cfv.mass(self.df_gen["P_projectile"].to_numpy())
            self.mass_target = Cfv.mass(self.df_gen["P_target"].to_numpy())
            self.mass_decay_N_parent = Cfv.mass(self.df_gen["P_decay_N_parent"].to_numpy())
            self.mass_recoil = Cfv.mass(self.df_gen["P_recoil"].to_numpy())
            self.mass_decay_N_daughter = Cfv.mass(self.df_gen["P_decay_N_daughter"].to_numpy())
            self.mass_decay_N_daughter[np.isnan(self.mass_decay_N_daughter)] = 0.0
            self.mass_decay_ell_minus = Cfv.mass(self.df_gen["P_decay_ell_minus"].to_numpy())
            self.mass_decay_ell_plus = Cfv.mass(self.df_gen["P_decay_ell_plus"].to_numpy())

            self.pvec_projectile = self.df_gen["P_projectile"].to_numpy()
            self.pvec_target = self.df_gen["P_target"].to_numpy()
            self.pvec_decay_N_parent = self.df_gen["P_decay_N_parent"].to_numpy()
            self.pvec_recoil = self.df_gen["P_recoil"].to_numpy()
            self.pvec_decay_N_daughter = self.df_gen["P_decay_N_daughter"].to_numpy()

            if self.decay_product == "e+e-" or self.decay_product == "mu+mu-":
                self.pvec_decay_ell_minus = self.df_gen["P_decay_ell_minus"].to_numpy()
                self.pvec_decay_ell_plus = self.df_gen["P_decay_ell_plus"].to_numpy()
            if self.decay_product == "photon":
                self.pvec_decay_photon = self.df_gen["P_decay_photon"].to_numpy()

            self.pvec_pos_decay = self.df_gen["pos_decay"].to_numpy()
            self.pvec_pos_scatt = self.df_gen["pos_scatt"].to_numpy()

            # string to be saved to file
            self.projectile_flavor = int(lp.nu_mu.pdgid)
            if self.decay_product == "e+e-":
                self.id_lepton_minus = int(lp.e_minus.pdgid)
                self.id_lepton_plus = int(lp.e_plus.pdgid)
            elif self.decay_product == "mu+mu-":
                self.id_lepton_minus = int(lp.mu_minus.pdgid)
                self.id_lepton_plus = int(lp.mu_plus.pdgid)
            elif self.decay_product == "photon":
                pass
            else:
                logger.warning(f"Decay product {self.decay_product} not recognized, assuming it to be e+e-.")
                self.id_lepton_minus = int(lp.e_minus.pdgid)
                self.id_lepton_plus = int(lp.e_plus.pdgid)

            # kinematics converted already
            self._kinematics_in_np_arrays = True

        # # sample size (# of events)
        # tot_events_to_print = len(df_gen.index)

        # if not 'pos_scatt' in df_gen.columns:
        # 	logger.debug("DEBUG: enforcing pos_scatt = 0")
        # 	df_gen['pos_scatt', '0'] = np.zeros((tot_events_to_print, 0))
        # 	df_gen['pos_scatt', '1'] = np.zeros((tot_events_to_print, 0))
        # 	df_gen['pos_scatt', '2'] = np.zeros((tot_events_to_print, 0))
        # 	df_gen['pos_scatt', '3'] = np.zeros((tot_events_to_print, 0))

        # if not 'pos_decay' in df_gen.columns:
        # 	logger.debug("DEBUG: enforcing pos_decay = pos_scatt")
        # 	df_gen['pos_decay', '0'] = df_gen['pos_scatt', '0']
        # 	df_gen['pos_decay', '1'] = df_gen['pos_scatt', '1']
        # 	df_gen['pos_decay', '2'] = df_gen['pos_scatt', '2']
        # 	df_gen['pos_decay', '3'] = df_gen['pos_scatt', '3']

        # # pre-computing some variables

        # mass_projectile = Cfv.mass(df_gen['P_projectile'].to_numpy())
        # mass_target = Cfv.mass(df_gen['P_target'].to_numpy())
        # mass_decay_N_parent = Cfv.mass(df_gen['P_decay_N_parent'].to_numpy())
        # mass_recoil = Cfv.mass(df_gen['P_recoil'].to_numpy())
        # mass_decay_N_daughter = Cfv.mass(df_gen['P_decay_N_daughter'].to_numpy())

        # pvec_projectile = df_gen['P_projectile'][['1','2','3']].to_numpy()
        # pvec_target = df_gen['P_target'][['1','2','3']].to_numpy()
        # pvec_decay_N_parent = df_gen['P_decay_N_parent'][['1','2','3']].to_numpy()
        # pvec_recoil = df_gen['P_recoil'][['1','2','3']].to_numpy()
        # pvec_decay_N_daughter = df_gen['P_decay_N_daughter'][['1','2','3']].to_numpy()
        # pvec_decay_ell_minus = df_gen['P_decay_ell_minus'][['1','2','3']].to_numpy()
        # pvec_decay_ell_plus = df_gen['P_decay_ell_plus'][['1','2','3']].to_numpy()

        # pvec_pos_decay = df_gen['pos_decay'][['1','2','3']].to_numpy()
        # pvec_pos_scatt = df_gen['pos_scatt'][['1','2','3']].to_numpy()

        # projectile_flavor = int(lp.nu_mu.pdgid)
        # if decay_product == 'e+e-':
        # 	id_lepton_minus = int(lp.e_minus.pdgid)
        # 	id_lepton_plus = int(lp.e_plus.pdgid)
        # elif decay_product == 'mu+mu-':
        # 	id_lepton_minus = int(lp.mu_minus.pdgid)
        # 	id_lepton_plus = int(lp.mu_plus.pdgid)
        # else:
        # 	id_lepton_minus = int(lp.e_minus.pdgid)
        # 	id_lepton_plus = int(lp.e_plus.pdgid)
        # 	logger.warning(f'Decay product {decay_product} not recognized, assuming it to be e+e-.')

    def print_events_to_hepevt(self, filename=None, hep_unweight=False, unweighted_hep_events=100):
        if not hep_unweight:
            logger.warning(
                "WARNING: HEPevt is not a lossless format -- you will lose the event weights. \
If you want to force-print weights, use the hepevt_legacy format instead. \
Otherwise, please set hep_unweight=True and set the desired number of unweighted events."
            )
        # HEPevt file name
        if filename:
            hep_path = filename
        else:
            hep_path = Path(f"{self.out_file_name}/HEPevt.dat").__str__()
        self._pyhepmc_printer(hep.WriterHEPEVT(hep_path), hep_unweight=hep_unweight, unweighted_hep_events=unweighted_hep_events)

    def print_events_to_hepmc2(self, filename=None, hep_unweight=False, unweighted_hep_events=100):
        # HEPevt file name
        if filename:
            hep_path = filename
        else:
            hep_path = Path(f"{self.out_file_name}/hep_ascii.hepmc2").__str__()
        self._pyhepmc_printer(hep.WriterAsciiHepMC2(hep_path), hep_unweight=hep_unweight, unweighted_hep_events=unweighted_hep_events)

    def print_events_to_hepmc3(self, filename=None, hep_unweight=False, unweighted_hep_events=100):
        # HEPevt file name
        if filename:
            hep_path = filename
        else:
            hep_path = Path(f"{self.out_file_name}/hep_ascii.hepmc3").__str__()
        self._pyhepmc_printer(hep.WriterAscii(hep_path), hep_unweight=hep_unweight, unweighted_hep_events=unweighted_hep_events)

    def _pyhepmc_printer(self, hep_writer, hep_unweight=False, unweighted_hep_events=100):
        """ Use pyhepmc to print events to standard HEP formats.

		Args:
			hep_writer (an instance of a hep writer bindings): of one of the following types: WriterHEPEVT, WriterAsciiHepMC2, WriterAscii
			hep_unweight (bool, optional): if true, unweight events. Defaults to False.
			unweighted_hep_events (int, optional): if hep_unweight is true, use this value to determine how many unweighted events to print. Defaults to 100.

		Events are printed using the numpy arrays that have obtained from the pandas dataframe. This speeds up the printing process
		Four momenta are printed following the convention:
			
			px py pz e pdgid status_code

			status_code is defined as:
				0 Not defined (null entry) Not a meaningful status
				1 Undecayed physical particle Recommended for all cases
				2 Decayed physical particle Recommended for all cases
				3 Documentation line Often used to indicate
				in/out particles in hard process
				4 Incoming beam particle Recommended for all cases
				5-10 Reserved for future standards Should not be used
				11-200 Generator-dependent For generator usage
				201- Simulation-dependent For simulation software usage
				

		"""

        # pre-compute kinematics with numpy for faster index access
        self._prepare_kinematics(hep_unweight=hep_unweight, unweighted_hep_events=unweighted_hep_events)

        # converting Lorentz order -- px, py, pz, E
        hep_order = [1, 2, 3, 0]

        ri = hep.GenRunInfo()
        from . import __version__

        ri.tools = [("DarkNews", __version__, "DarkNews upscattering engine")]

        # name of the weights -- w_event_rate, w_flux_avg_xsec, w_decay_rate_0 (, w_decay_rate_1)
        df_weight_names = [name for name in list(self.df_gen.columns.levels[0]) if "w" in name]
        ri.weight_names = df_weight_names

        # for i in range(df.index[-1]):
        for i in range(unweighted_hep_events):
            evt = hep.GenEvent(hep.Units.GEV, hep.Units.CM)
            evt.event_number = i

            #### Upscattering process
            p1 = hep.GenParticle(self.pvec_projectile[i, hep_order], self.projectile_flavor, 4)
            p1.generated_mass = self.mass_projectile[i]
            evt.add_particle(p1)

            p2 = hep.GenParticle(self.pvec_target[i, hep_order], int(self.df_gen["target_pdgid", ""].to_numpy()[i]), 4)
            p2.generated_mass = self.mass_target[i]
            evt.add_particle(p2)

            p3 = hep.GenParticle(self.pvec_decay_N_parent[i, hep_order], int(pdg.neutrino5.pdgid), 2)
            p3.generated_mass = self.mass_decay_N_parent[i]
            evt.add_particle(p3)

            p4 = hep.GenParticle(self.pvec_recoil[i, hep_order], int(self.df_gen["target_pdgid", ""].to_numpy()[i]), 2)
            p4.generated_mass = self.mass_recoil[i]
            evt.add_particle(p4)

            v1 = hep.GenVertex((self.pvec_pos_scatt[i, hep_order]))
            v1.add_particle_in(p1)
            v1.add_particle_in(p2)
            v1.add_particle_out(p3)
            v1.add_particle_out(p4)

            #### Decay of HNL
            pnu = hep.GenParticle(self.pvec_decay_N_daughter[i, hep_order], int(pdg.neutrino4.pdgid), 1)
            pnu.generated_mass = self.mass_decay_N_daughter[i]
            evt.add_particle(pnu)
            v2 = hep.GenVertex((self.pvec_pos_decay[i, hep_order]))
            v2.add_particle_in(p3)
            v2.add_particle_out(pnu)

            if self.decay_product == "e+e-" or self.decay_product == "mu+mu-":
                pep = hep.GenParticle(self.pvec_decay_ell_plus[i, hep_order], int(self.id_lepton_plus), 1)
                pep.generated_mass = self.mass_decay_ell_plus[i]
                evt.add_particle(pep)
                pem = hep.GenParticle(self.pvec_decay_ell_minus[i, hep_order], int(self.id_lepton_minus), 1)
                pem.generated_mass = self.mass_decay_ell_minus[i]
                evt.add_particle(pem)

                v2.add_particle_out(pep)
                v2.add_particle_out(pem)

            elif self.decay_product == "photon":
                pphoton = hep.GenParticle(self.pvec_decay_photon[i, hep_order], int(lp.photon.pdgid), 1)
                pphoton.generated_mass = 0.0

                v2.add_particle_out(pphoton)

            # now add the scattering and decay vertices
            evt.add_vertex(v1)
            evt.add_vertex(v2)

            evt.run_info = ri
            if not hep_unweight:
                evt.weights = [self.df_gen[name, ""].to_numpy()[i] for name in df_weight_names]

            # write event to file using the chosen hep writer (HEPevt, hepmc2 or 3)
            hep_writer.write(evt)

    def print_events_to_hepevt_legacy(self, filename=None, hep_unweight=False, unweighted_hep_events=100):
        """
			Print events to HEPevt format.

				The file start with the total number of events:
					
					'tot_events_to_print'

				On a new line, each event starts with a brief description of the event:
					
					'event_number number_of_particles (event_weight if it exists)'

				On a new line, a new particle and its properties are added. Using string concatenation, 
				we print each particle as follows:
				
				(	
					f'0 '				  	# ignored = 0 or tracked = 1
					f' {i}'				  	# particle PDG number
					f' 0 0 0 0' 		  	# ????? (parentage)
					f' {*list([px,py,pz])}'	# particle px py pz momenta
					f' {E}' 				# particle energy
					f' {m}'				  	# particle mass
					f' {*list([x,y,z])}' 	# spatial x y z coords
					f' {t}' 				# time coord
					'\n'
				)

				The last two steps repeat until the EOF.

		"""

        # pre-compute kinematics with numpy for faster index access
        self._prepare_kinematics(hep_unweight=hep_unweight, unweighted_hep_events=unweighted_hep_events)

        logger.info("Printing events to HEPevt using legacy format...")
        lines = []


        range_of_evts = self.df_gen.index
        # loop over events
        for i in range_of_evts:
            # no particles & event id
            if hep_unweight:
                lines.append(f"{i} 7\n")
            else:
                lines.append(f"{i} 7 {self.df_gen['w_event_rate',''].to_numpy()[i]:.8E}\n")

            # scattering inital states
            lines.append(
                (  # Projectile
                    f"0 "
                    f" {self.projectile_flavor}"
                    f" 0 0 0 0"
                    f" {print_in_order(self.pvec_projectile[i])}"
                    f" {self.df_gen['P_projectile','0'].to_numpy()[i]:.8E}"
                    f" {self.mass_projectile[i]:.8E}"
                    f" {print_in_order(self.pvec_pos_scatt[i])}"
                    f" {self.df_gen['pos_scatt','0'].to_numpy()[i]:.8E}"
                    "\n"
                )
            )
            
            # Only print targets and HNLs if not sparse
            if not self.sparse:
                lines.append(
                    (  # Target
                        f"0 "
                        f" {int(self.df_gen['target_pdgid',''].to_numpy()[i])}"
                        f" 0 0 0 0"
                        f" {print_in_order(self.pvec_target[i])}"
                        f" {self.df_gen['P_target','0'].to_numpy()[i]:.8E}"
                        f" {self.mass_target[i]:.8E}"
                        f" {print_in_order(self.pvec_pos_scatt[i])}"
                        f" {self.df_gen['pos_scatt','0'].to_numpy()[i]:.8E}"
                        "\n"
                    )
                )

                # scatter final products
                lines.append(
                    (  # HNL produced
                        f"0 "
                        f" {int(pdg.neutrino5.pdgid)}"
                        f" 0 0 0 0"
                        f" {print_in_order(self.pvec_decay_N_parent[i])}"
                        f" {self.df_gen['P_decay_N_parent','0'].to_numpy()[i]:.8E}"
                        f" {self.mass_decay_N_parent[i]:.8E}"
                        f" {print_in_order(self.pvec_pos_scatt[i])}"
                        f" {self.df_gen['pos_scatt','0'].to_numpy()[i]:.8E}"
                        "\n"
                    )
                )

                lines.append(
                    (  # recoiled target
                        f"0 "
                        f" {int(self.df_gen['target_pdgid',''].to_numpy()[i])}"
                        f" 0 0 0 0"
                        f" {print_in_order(self.pvec_recoil[i])}"
                        f" {self.df_gen['P_recoil','0'].to_numpy()[i]:.8E}"
                        f" {self.mass_recoil[i]:.8E}"
                        f" {print_in_order(self.pvec_pos_scatt[i])}"
                        f" {self.df_gen['pos_scatt','0'].to_numpy()[i]:.8E}"
                        "\n"
                    )
                )

                # decay final products
                lines.append(
                    (  # daughter neutrino/HNL
                        f"0 "
                        f" {int(pdg.nulight.pdgid)}"
                        f" 0 0 0 0"
                        f" {print_in_order(self.pvec_decay_N_daughter[i])}"
                        f" {self.df_gen['P_decay_N_daughter','0'].to_numpy()[i]:.8E}"
                        f" {self.mass_decay_N_daughter[i]:.8E}"
                        f" {print_in_order(self.pvec_pos_decay[i])}"
                        f" {self.df_gen['pos_decay','0'].to_numpy()[i]:.8E}"
                        "\n"
                    )
                )

            lines.append(
                (  # ell-
                    f"1 "
                    f" {self.id_lepton_minus}"
                    f" 0 0 0 0"
                    f" {print_in_order(self.pvec_decay_ell_minus[i])}"
                    f" {self.df_gen['P_decay_ell_minus','0'].to_numpy()[i]:.8E}"
                    f" {const.m_e:.8E}"
                    f" {print_in_order(self.pvec_pos_decay[i])}"
                    f" {self.df_gen['pos_decay','0'].to_numpy()[i]:.8E}"
                    "\n"
                )
            )

            lines.append(
                (  # ell+
                    f"1 "
                    f" {self.id_lepton_plus}"
                    f" 0 0 0 0"
                    f" {print_in_order(self.pvec_decay_ell_plus[i])}"
                    f" {self.df_gen['P_decay_ell_plus','0'].to_numpy()[i]:.8E}"
                    f" {const.m_e:.8E}"
                    f" {print_in_order(self.pvec_pos_decay[i])}"
                    f" {self.df_gen['pos_decay','0'].to_numpy()[i]:.8E}"
                    "\n"
                )
            )

        # string to be saved to file
        hepevt_string = ""
        hepevt_string += f"{self.tot_events_to_print}\n"

        # HEPevt file name
        if filename:
            hepevt_file_name = Path(filename).__str__()
        else:
            hepevt_file_name = Path(f"{self.out_file_name}/HEPevt.dat").__str__()

        f = open(hepevt_file_name, "w+")
        # print events
        f.write(hepevt_string + "".join(lines))
        f.close()

        prettyprinter.info(f"HEPevt events saved to file successfully:\n{hepevt_file_name}")



import numpy as np
from scipy import interpolate
from pathlib import Path
import os.path
import os
try:
    import importlib.resources as resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as resources


from . import logger, prettyprinter
from . import pdg
from . import geom
from . import const
from .nuclear_tools import NuclearTarget
from .AssignmentParser import AssignmentParser


class Detector:
    """
    Detector is a collection of necessary variables for cross-section and event rate
    calculations, e.g., energy range, target type, weight, and POTs for exposure.

    It provides the `get_flux_func` which creates a spline interpolation of the
    input neutrino flux.

    Args:
        exp_module (name):  name of experiment corresponding to a .txt file
                            in the path `DarkNews/include/detector/exp_module.txt`.
                            the file should contain parameters specific for
                            that experiment if it is user defined.

    Raises:
        FileNotFoundError: if no detector file was found
        KeyError: if a required field is not specified in the detector file

    """
    KEYWORDS = {
        "dune_nd_fhc": "dune_nd_fhc.txt",
        "dune_nd_rhc": "dune_nd_rhc.txt",
        "microboone": "microboone.txt",
        "minerva_le_fhc": "minerva_le_fhc.txt",
        "minerva_me_fhc": "minerva_me_fhc.txt",
        "minerva_me_rhc": "minerva_me_fhc.txt",
        "miniboone_fhc": "miniboone_fhc.txt",
        "miniboone_rhc": "miniboone_rhc.txt",
        "minos_le_fhc": "minos_le_fhc.txt",
        "nd280_fhc": "nd280_fhc.txt",
        "nova_le_fhc": "nova_le_fhc.txt",
        "fasernu": "fasernu.txt",
        "nutev_fhc": "nutev_fhc.txt",
    }

    def __init__(self, experiment):
        parser = AssignmentParser({})
        DET_MODULE = "DarkNews.include.detectors"
        kwargs = {"encoding": "utf8"}
        try:
            # experiment is initially interpreted as a path to a local file
            experiment_file = experiment
            parser.parse_file(file=experiment_file, comments="#")
        except (OSError, IOError, FileNotFoundError) as err:
            # if no file is found, then it is interpreted as a keyword for a pre-defined experiment
            if experiment in self.KEYWORDS:
                with resources.open_text(DET_MODULE, self.KEYWORDS[experiment], **kwargs) as f:
                    parser.parse_file(file=f, comments="#")
            else:
                raise err

        params = parser.parameters
        try:
            self.NAME = params["name"]
            self.FLUXFILE = params["fluxfile"]
            self.FLUX_NORM = params["flux_norm"]
            self.ERANGE = params["erange"]

            # Detector targets
            self.NUCLEAR_TARGETS = [NuclearTarget(target) for target in params["nuclear_targets"]]
            self.POTS = params["POTs"]
            self.FIDUCIAL_MASS_PER_TARGET = params["fiducial_mass_per_target"]

        except KeyError as err:
            if err.args[0] in [
                "name",
                "fluxfile",
                "flux_norm",
                "erange",
                "nuclear_targets",
                "fiducial_mass_per_target",
                "POTs",
            ]:
                # check that the error comes from reading the file and not elsewhere
                raise KeyError(f"No field '{err.args[0]}' specified in the the experiment configuration file '{experiment}'.".format(err.args[0], experiment))
            raise err

        # total number of targets
        self.NUMBER_OF_TARGETS = {}
        for fid_mass, target in zip(self.FIDUCIAL_MASS_PER_TARGET, self.NUCLEAR_TARGETS):
            self.NUMBER_OF_TARGETS[f"{target.name}"] = fid_mass * const.t_to_GeV / (target.mass)

        # load neutrino fluxes: first try with path relative to experiment file, if error try with path from original config files
        try:
            exp_dir = os.path.dirname(experiment_file)
            _enu, *_fluxes = np.genfromtxt(f"{exp_dir}/{self.FLUXFILE}", unpack=True)
        except (OSError, FileNotFoundError, TypeError):
            try:
                file = resources.open_text("DarkNews.include.fluxes", self.FLUXFILE)
                _enu, *_fluxes = np.genfromtxt(file, unpack=True)
            except FileNotFoundError:
                raise FileNotFoundError(f"Fluxes file {self.FLUXFILE} not found in current experiment file path nor in config file path.")
            
        self.FLUX_FUNCTIONS = 6 * [[]]
        for i in range(len(_fluxes)):
            self.FLUX_FUNCTIONS[i] = interpolate.interp1d(_enu, _fluxes[i] * self.FLUX_NORM, fill_value=0.0, bounds_error=False)

        prettyprinter.info(
            f"""Experiment:
\t{self.NAME}
\tfluxfile loaded: {self.FLUXFILE}
\tPOT: {self.POTS}
\tnuclear targets: {[n.name for n in self.NUCLEAR_TARGETS]}
\tfiducial mass: {self.FIDUCIAL_MASS_PER_TARGET} tonnes"""
        )

    def neutrino_flux(self, projectile):
        _flux_index = pdg.get_doublet(projectile) + 3 * pdg.is_antiparticle(projectile)
        return self.FLUX_FUNCTIONS[_flux_index]

    def set_geometry(self):
        if "microboone" in self.NAME.lower():
            self.place_scatters = geom.microboone_geometry
        elif "miniboone" in self.NAME.lower():
            self.place_scatters = geom.miniboone_geometry
        else:
            self.place_scatters = geom.point_geometry



import numpy as np
import vegas as vg
from collections import OrderedDict

from DarkNews import logger, prettyprinter

from DarkNews import Cfourvec as Cfv
from DarkNews import const
from DarkNews import phase_space
from DarkNews import decay_rates as dr
from DarkNews import amplitudes as amps


class UpscatteringXsec(vg.BatchIntegrand):
    def __init__(self, dim, Enu, ups_case, diagram="total"):
        """
        Vegas integrand for diff cross section for upscattering

        Args:
                dim (int): integration dimensions
                Enu (float): neutrino energy to be considered
                ups_case (DarkNews.processes.UpscatteringProcess): the upscattering class of DarkNews
                diagram (str, optional): _description_. Defaults to 'total'.

        Raises:
                ValueError: if cannot find diagrams to be computed
        """
        self.dim = dim
        self.Enu = Enu
        self.ups_case = ups_case
        self.diagram = diagram
        if not isinstance(self.diagram, str):
            logger.error(f"ERROR. Cannot calculate total cross section for more than one diagram at a time. Passed diagram={self.diagram}.")
            raise ValueError

        # Enforce Q2 range
        if self.ups_case.scattering_regime == "coherent":
            self.QMIN = 0  # GeV
            self.QMAX = 2  # GeV
        elif self.ups_case.scattering_regime in ["p-el", "n-el"]:
            self.QMIN = 0  # GeV
            self.QMAX = 5  # GeV
        elif self.ups_case.scattering_regime in ["DIS"]:
            self.QMIN = 2  # GeV
            self.QMAX = np.inf  # GeV
        else:
            self.QMIN = 0  # GeV
            self.QMAX = np.inf

        # Find the normalization factor
        self.norm = {}
        self.norm["diff_xsec"] = 1
        # normalize integrand with an initial throw
        _throw = self.__call__(np.random.rand(self.dim, 20000), np.ones((self.dim, 20000)))
        for key, val in _throw.items():
            self.norm[key] = np.mean(val)
            # cannot normalize zero integrand
            if self.norm[key] == 0.0:
                logger.warning("Warning: mean of integrand is zero. Vegas may break.")
                self.norm[key] = 1

    def __call__(self, x, jac):

        ups_case = self.ups_case

        ##############################################
        # Upscattering Kinematics
        Enu = self.Enu
        M = ups_case.target.mass

        Q2lmin = np.log(phase_space.upscattering_Q2min(Enu, ups_case.m_ups, M))
        Q2lmax = np.log(np.minimum(phase_space.upscattering_Q2max(Enu, ups_case.m_ups, M), self.QMAX ** 2))

        Q2l = (Q2lmax - Q2lmin) * x[:, 0] + Q2lmin
        Q2 = np.exp(Q2l)

        s = M ** 2 + 2 * Enu * M  # massless projectile
        t = -Q2
        u = 2 * M ** 2 + ups_case.m_ups ** 2 - s - t  # massless projectile

        ##############################################
        # Upscattering amplitude squared (spin summed -- not averaged)
        diff_xsec = amps.upscattering_dxsec_dQ2([s, t, u], self.ups_case, diagrams=[self.diagram])
        if type(diff_xsec) is dict:
            diff_xsec = np.sum([diff_xsec[diagram] for diagram in diff_xsec.keys()])
        # hypercube jacobian (vegas hypercube --> physical limits) transformation
        hypercube_jacobian = (Q2lmax - Q2lmin) * np.exp(Q2l)
        diff_xsec *= hypercube_jacobian

        ##############################################
        # return all differential quantities of interest
        self.int_dic = OrderedDict()
        self.int_dic["diff_xsec"] = diff_xsec
        ##############################################
        # normalization
        self.int_dic["diff_xsec"] /= self.norm["diff_xsec"]

        return self.int_dic


class UpscatteringHNLDecay(vg.BatchIntegrand):
    def __init__(self, dim, Emin, Emax, MC_case):
        """
        Vegas integrand for the process of upscattering with subsequent decays.
        Scattering diff xsec and diff decay rates are integrated simultaneously.

        Args:
                dim (int): _description_
                Emin (float): min neutrino energy to integrate flux
                Emax (float): max neutrino energy to integrate flux
                MC_case (DarkNews.MC.MC_events): the main Monte-Carlo class of DarkNews

        Raises:
                ValueError: if cannot find what decay process to consider
        """
        self.dim = dim
        self.Emax = Emax
        self.Emin = Emin
        self.MC_case = MC_case

        self.norm = {}
        self.norm["diff_event_rate"] = 1
        self.norm["diff_flux_avg_xsec"] = 1
        if self.MC_case.decays_to_dilepton:
            if self.MC_case.decay_case.on_shell:
                self.norm["diff_decay_rate_0"] = 1
                self.norm["diff_decay_rate_1"] = 1
            elif self.MC_case.decay_case.off_shell:
                self.norm["diff_decay_rate_0"] = 1
        elif self.MC_case.decays_to_singlephoton:
            self.norm["diff_decay_rate_0"] = 1
        else:
            logger.error("ERROR: Could not determine decay process in vegas integral.")
            raise ValueError

        # normalize integrand with an initial throw
        logger.debug("Throwing an initial 10000 random points to find the normalization")
        _throw = self.__call__(np.random.rand(self.dim, 10_000), np.ones((self.dim, 10_000)))
        logger.debug("Throwing successful")
        for key, val in _throw.items():
            self.norm[key] = np.mean(val)
            # cannot normalize zero integrand
            if self.norm[key] == 0:
                self.norm[key] = 1

    def __call__(self, x, jac):

        self.int_dic = OrderedDict()

        ups_case = self.MC_case.ups_case
        decay_case = self.MC_case.decay_case

        M = ups_case.target.mass
        m_parent = ups_case.m_ups
        m_daughter = decay_case.m_daughter
        mzprime = ups_case.mzprime

        i_var = 0
        # neutrino energy
        Enu = (self.Emax - self.Emin) * x[:, i_var] + self.Emin
        i_var += 1

        ##############################################
        # Upscattering differential cross section (spin averaged)

        Q2lmin = np.log(phase_space.upscattering_Q2min(Enu, m_parent, M))
        Q2lmax = np.log(phase_space.upscattering_Q2max(Enu, m_parent, M))

        Q2l = (Q2lmax - Q2lmin) * x[:, i_var] + Q2lmin
        i_var += 1

        Q2 = np.exp(Q2l)
        s_scatt = M ** 2 + 2 * Enu * M  # massless projectile
        t_scatt = -Q2
        u_scatt = 2 * M ** 2 + m_parent ** 2 - s_scatt + Q2  # massless projectile

        diff_xsec = amps.upscattering_dxsec_dQ2([s_scatt, t_scatt, u_scatt], ups_case)

        # hypercube jacobian (vegas hypercube --> physical limits) transformation
        hypercube_jacobian = (Q2lmax - Q2lmin) * np.exp(Q2l) * (self.Emax - self.Emin)
        diff_xsec *= hypercube_jacobian

        self.int_dic["diff_event_rate"] = diff_xsec * self.MC_case.flux(Enu)
        self.int_dic["diff_flux_avg_xsec"] = diff_xsec * self.MC_case.flux(Enu)

        ##############################################
        if self.MC_case.decays_to_dilepton:

            if decay_case.on_shell:
                ##############################################
                # decay nu_parent -> nu_daughter mediator

                # angle between nu_daughter and z axis
                cost = -1.0 + (2.0) * x[:, i_var]
                i_var += 1

                # params.UD4**2 * (params.Ue4**2 + params.Umu4**2 + params.Utau4**2)*params.gD**2
                self.int_dic["diff_decay_rate_0"] = dr.diff_gamma_Ni_to_Nj_V(
                    cost=cost, vertex_ij=np.sqrt(decay_case.Dih), mi=m_parent, mj=m_daughter, mV=mzprime, HNLtype=decay_case.HNLtype, h=decay_case.h_parent,
                )
                self.int_dic["diff_decay_rate_0"] *= 2  # hypercube jacobian

                ##############################################
                # mediator decay M --> ell+ ell-
                self.int_dic["diff_decay_rate_1"] = dr.gamma_V_to_ell_ell(vertex=decay_case.TheoryModel.deV, mV=mzprime, m_ell=decay_case.mm) * np.full_like(
                    self.int_dic["diff_decay_rate_0"], 1.0
                )

            elif decay_case.off_shell:
                ##############################################
                # decay nu_parent -> nu_daughter ell+ ell-

                m1 = decay_case.m_parent
                m2 = decay_case.mm
                m3 = decay_case.mp
                m4 = decay_case.m_daughter
                masses = np.array([m1, m2, m3, m4])

                # limits
                tmax = phase_space.three_body_tmax(*masses)
                tmin = phase_space.three_body_tmin(*masses)
                t = (tmax - tmin) * x[:, i_var] + tmin
                i_var += 1

                umax = phase_space.three_body_umax(*masses, t)
                umin = phase_space.three_body_umin(*masses, t)
                u = (umax - umin) * x[:, i_var] + umin
                i_var += 1

                v = np.sum(masses ** 2) - u - t

                c3 = (2.0) * x[:, i_var] - 1.0
                i_var += 1
                phi34 = (2.0 * np.pi) * x[:, i_var]
                i_var += 1

                dgamma = dr.diff_gamma_Ni_to_Nj_ell_ell([t, u, v, c3, phi34], decay_case)

                # hypercube jacobian (vegas hypercube --> physical limits) transformation
                dgamma *= tmax - tmin
                dgamma *= umax - umin
                dgamma *= 2  # c3
                dgamma *= 2 * np.pi  # phi34
                self.int_dic["diff_decay_rate_0"] = dgamma

        elif self.MC_case.decays_to_singlephoton:
            ##############################################
            # decay nu_parent -> nu_daughter gamma

            # angle between nu_daughter and z axis
            cost = -1.0 + (2.0) * x[:, i_var]
            i_var += 1

            self.int_dic["diff_decay_rate_0"] = dr.diff_gamma_Ni_to_Nj_gamma(
                cost=cost, vertex_ij=decay_case.Tih, mi=m_parent, mj=m_daughter, HNLtype=decay_case.HNLtype, h=decay_case.h_parent,
            )

            # hypercube jacobian (vegas hypercube --> physical limits) transformation
            self.int_dic["diff_decay_rate_0"] *= 2

        else:
            logger.error("ERROR: Could not determine decay process in vegas integral.")
            raise ValueError

        ##############################################
        # storing normalization factor that guarantees that integrands are O(1) numbers
        # loop over decay processes
        for decay_step in (k for k in self.int_dic.keys() if "decay_rate" in k):
            # multiply differential event rate by dGamma_i/dPS
            self.int_dic["diff_event_rate"] *= self.int_dic[decay_step]

        # normalize integrands to be O(1)
        for k in self.norm.keys():
            self.int_dic[k] /= self.norm[k]
        logger.debug(f"Normalization factors in integrand: {self.norm}.")

        ##############################################
        # flat direction jacobian (undoing the vegas adaptation in flat directions of factors of the full integrands)
        # loop over decay processes
        for decay_step in (k for k in self.int_dic.keys() if "decay_rate" in k):
            self.int_dic[decay_step] /= jac[:, 0] * jac[:, 1]
        decay_directions = range(2, self.dim)
        for d in decay_directions:
            self.int_dic["diff_flux_avg_xsec"] /= jac[:, d]

        # return all differential quantities of interest
        return self.int_dic


def get_momenta_from_vegas_samples(vsamples=None, MC_case=None):
    """
    Construct the four momenta of all particles in the upscattering+decay process from the
    vegas weights.

    Args:
            vsamples (np.ndarray, optional): integration samples obtained from vegas
                            as hypercube coordinates. Always in the interval [0,1].

            MC_case (DarkNews.MC.MC_events): the main Monte-Carlo class of DarkNews

    Returns:
            dict: each key corresponds to a set of four momenta for a given particle involved,
                    so the values are 2D np.ndarrays with each row a different event and each column a different
                    four momentum component. Contains also the weights.
    """

    four_momenta = {}

    ########################
    ### upscattering
    # Ni(k1) target(k2) -->  Nj(k3) target(k4)
    mh = MC_case.ups_case.m_ups
    MA = MC_case.ups_case.MA

    # energy of projectile
    Eprojectile = (MC_case.EMAX - MC_case.EMIN) * vsamples[0] + MC_case.EMIN
    scatter_samples = {"Eprojectile": Eprojectile, "unit_Q2": vsamples[1]}
    masses_scatter = {
        "m1": 0.0,  # nu_projectile
        "m2": MA,  # target
        "m3": mh,  # nu_upscattered
        "m4": MA,  # final target
    }

    P1LAB, P2LAB, P3LAB, P4LAB = phase_space.two_to_two_scatter(scatter_samples, **masses_scatter)

    # N boost parameters
    boost_scattered_N = {
        "EP_LAB": P3LAB.T[0],
        "costP_LAB": Cfv.get_cosTheta(P3LAB),
        "phiP_LAB": np.arctan2(P3LAB.T[2], P3LAB.T[1]),
    }

    four_momenta["P_projectile"] = P1LAB
    four_momenta["P_target"] = P2LAB
    four_momenta["P_recoil"] = P4LAB

    #######################
    # DECAY PROCESSES

    if MC_case.decays_to_dilepton:

        mf = MC_case.decay_case.m_daughter
        mm = MC_case.decay_case.mm
        mp = MC_case.decay_case.mm
        mzprime = MC_case.decay_case.mzprime

        if MC_case.decay_case.on_shell:

            ########################
            ### HNL decay
            N_decay_samples = {"unit_cost": np.array(vsamples[2])}
            # Ni (k1) --> Nj (k2)  Z' (k3)
            masses_decay = {
                "m1": mh,  # Ni
                "m2": mf,  # Nj
                "m3": mzprime,  # Z'
            }
            # Phnl, Phnl_daughter, Pz'
            P1LAB_decay, P2LAB_decay, P3LAB_decay = phase_space.two_body_decay(N_decay_samples, boost=boost_scattered_N, **masses_decay)

            # Z' boost parameters
            boost_Z = {
                "EP_LAB": P3LAB_decay.T[0],
                "costP_LAB": Cfv.get_cosTheta(P3LAB_decay),
                "phiP_LAB": np.arctan2(P3LAB_decay.T[2], P3LAB_decay.T[1]),
            }

            ########################
            ### Z' decay
            Z_decay_samples = {}  # all uniform
            # Z'(k1) --> ell- (k2)  ell+ (k3)
            masses_decay = {
                "m1": mzprime,  # Ni
                "m2": mp,  # \ell+
                "m3": mm,  # \ell-
            }
            # PZ', pe-, pe+
            P1LAB_decayZ, P2LAB_decayZ, P3LAB_decayZ = phase_space.two_body_decay(Z_decay_samples, boost=boost_Z, **masses_decay)

            four_momenta["P_decay_N_parent"] = P1LAB_decay
            four_momenta["P_decay_N_daughter"] = P2LAB_decay
            four_momenta["P_decay_ell_minus"] = P2LAB_decayZ
            four_momenta["P_decay_ell_plus"] = P3LAB_decayZ

        elif MC_case.decay_case.off_shell:

            ########################
            # HNL decay
            N_decay_samples = {
                "unit_t": vsamples[2],
                "unit_u": vsamples[3],
                "unit_c3": vsamples[4],
                "unit_phi34": vsamples[5],
            }

            # Ni (k1) --> ell-(k2)  ell+(k3)  Nj(k4)
            masses_decay = {
                "m1": mh,  # Ni
                "m2": mm,  # ell-
                "m3": mp,  # ell+
                "m4": mf,
            }  # Nj
            # Phnl, pe-, pe+, pnu
            (P1LAB_decay, P2LAB_decay, P3LAB_decay, P4LAB_decay,) = phase_space.three_body_decay(N_decay_samples, boost=boost_scattered_N, **masses_decay)

            four_momenta["P_decay_N_parent"] = P1LAB_decay
            four_momenta["P_decay_ell_minus"] = P2LAB_decay
            four_momenta["P_decay_ell_plus"] = P3LAB_decay
            four_momenta["P_decay_N_daughter"] = P4LAB_decay

    elif MC_case.decays_to_singlephoton:

        mf = MC_case.decay_case.m_daughter

        ########################
        ### HNL decay
        N_decay_samples = {"unit_cost": np.array(vsamples[2])}
        # Ni (k1) --> Nj (k2)  gamma (k3)
        masses_decay = {
            "m1": mh,  # Ni
            "m2": mf,  # Nj
            "m3": 0.0,  # gamma
        }
        # Phnl, Phnl', Pgamma
        P1LAB_decay, P2LAB_decay, P3LAB_decay = phase_space.two_body_decay(N_decay_samples, boost=boost_scattered_N, **masses_decay)

        four_momenta["P_decay_N_parent"] = P1LAB_decay
        four_momenta["P_decay_N_daughter"] = P2LAB_decay
        four_momenta["P_decay_photon"] = P3LAB_decay

    return four_momenta



import numpy as np
import math

from DarkNews import logger, prettyprinter

from particle import literals as lp

from DarkNews import const
from DarkNews import pdg
from DarkNews import decay_rates as dr


class HNLModel:
    def __init__(self, model_file=None, **model_params):
        """ HNL model class for models with HNLs + additional new physics

        Args:
            model_file (string, optional): The filename of the model file to load model parameters. Defaults to None.

        """
        self.model_file = model_file
        self.name = "my_model"

        # DEFAULTS
        self.m4 = 0.150
        self.m5 = None
        self.m6 = None
        self.HNLtype = "dirac"

        self.mzprime = 1.25
        self.mhprime = None
        
        self.mu_tr_e4 = 0.0
        self.mu_tr_e5 = 0.0
        self.mu_tr_e6 = 0.0
        self.mu_tr_mu4 = 0.0
        self.mu_tr_mu5 = 0.0
        self.mu_tr_mu6 = 0.0
        self.mu_tr_tau4 = 0.0
        self.mu_tr_tau5 = 0.0
        self.mu_tr_tau6 = 0.0
        self.mu_tr_44 = 0.0
        self.mu_tr_45 = 0.0
        self.mu_tr_46 = 0.0
        self.mu_tr_55 = 0.0
        self.mu_tr_56 = 0.0
        self.mu_tr_66 = 0.0

        self.s_e4 = 0.0
        self.s_e5 = 0.0
        self.s_e6 = 0.0
        self.s_mu4 = 0.0
        self.s_mu5 = 0.0
        self.s_mu6 = 0.0
        self.s_tau4 = 0.0
        self.s_tau5 = 0.0
        self.s_tau6 = 0.0
        self.s_44 = 0.0
        self.s_45 = 0.0
        self.s_46 = 0.0
        self.s_55 = 0.0
        self.s_56 = 0.0
        self.s_66 = 0.0

        # Initilize nucleon couplings. These will be filled with the quark combination, which is what is actually set by the user
        self.cprotonV = None
        self.cprotonA = None
        self.cneutronV = None
        self.cneutronA = None
        self.dprotonV = None
        self.dprotonA = None
        self.dneutronV = None
        self.dneutronA = None
        self.dSproton = None
        self.dSneutron = None
        self.dPproton = None
        self.dPneutron = None

        # Now assign parameters based on user input
        self.__dict__.update(model_params)

    def _initialize_spectrum(self):

        # Initialize spectrum
        self.nu_spectrum = [lp.nu_e, lp.nu_mu, lp.nu_tau]
        self._spectrum = ""
        self.hnl_masses = np.empty(0)

        if self.m4 is not None:
            self.hnl_masses = np.append(self.m4, self.hnl_masses)
            self.neutrino4 = pdg.neutrino4
            self.neutrino4.mass = self.m4
        if self.m5 is not None:
            self.hnl_masses = np.append(self.m5, self.hnl_masses)
            self.neutrino5 = pdg.neutrino5
            self.neutrino5.mass = self.m5
        if self.m6 is not None:
            self.hnl_masses = np.append(self.m6, self.hnl_masses)
            self.neutrino6 = pdg.neutrino6
            self.neutrino6.mass = self.m6

        # define new HNL particles and pass the masses in MeV units (default units for Particle...)
        for i in range(len(self.hnl_masses)):

            # PDGID  =  59(particle spin code: 0-scalar 1-fermion 2-vector)(generation number)
            # GeV units in particle module!
            hnl = pdg.new_particle(name=f"N{4+i}", pdgid=5914 + i, latex_name=f"N_{{{4+i}}}", mass=self.hnl_masses[i],)
            setattr(self, f"neutrino{4+i}", hnl)
            self.nu_spectrum.append(getattr(self, f"neutrino{4+i}"))

        self.HNL_spectrum = self.nu_spectrum[3:]
        self.n_nus = len(self.nu_spectrum)
        self.n_HNLs = len(self.HNL_spectrum)
        self._spectrum += f"\n\t{self.n_HNLs} {self.HNLtype} heavy neutrino(s)."

    def _update_spectrum(self):

        # mass mixing between Z' and Z
        # NOT YET FUNCTIONAL
        self.is_mass_mixed = False  # (self.epsilonZ != 0.0)

        self.has_Zboson_coupling = np.any(self.c_aj[3:, :] != 0)
        if self.has_Zboson_coupling:
            self._spectrum += f"\n\t{np.sum(self.c_aj[3:,:]!=0)} non-zero Z boson coupling(s) beyond the SM."

        self.has_vector_coupling = np.any(self.d_aj != 0)
        if self.has_vector_coupling:
            # dark photon
            self.zprime = pdg.new_particle(name="zprime", pdgid=5921, mass=self.mzprime * 1e3, latex_name="Z^\\prime",)
            self._spectrum += f"\n\t{np.sum(self.d_aj!=0)} non-zero Z'-neutrino coupling(s)."

        self.has_scalar_coupling = np.any(self.s_aj != 0)
        if self.has_scalar_coupling:
            # dark scalar
            self.hprime = pdg.new_particle(name="hprime", pdgid=5901, mass=self.mhprime * 1e3, latex_name="h^\\prime",)
            self._spectrum += f"\n\t{np.sum(self.s_aj!=0)} non-zero h'-neutrino coupling(s)."

        self.has_TMM = np.any(self.t_aj != 0)
        if self.has_TMM:
            self._spectrum += f"\n\t{np.sum(self.t_aj!=0)} non-zero transition magnetic moment(s)."


class GenericHNLModel(HNLModel):
    def __init__(self, model_file=None, **model_params):
        super().__init__(model_file, model_params=model_params)

        # set defaults
        self.c_e4 = 0.0
        self.c_e5 = 0.0
        self.c_e6 = 0.0
        self.c_mu4 = 0.0
        self.c_mu5 = 0.0
        self.c_mu6 = 0.0
        self.c_tau4 = 0.0
        self.c_tau5 = 0.0
        self.c_tau6 = 0.0
        self.c_44 = 0.0
        self.c_45 = 0.0
        self.c_46 = 0.0
        self.c_55 = 0.0
        self.c_56 = 0.0
        self.c_66 = 0.0

        self.d_e4 = 0.0
        self.d_e5 = 0.0
        self.d_e6 = 0.0
        self.d_mu4 = 0.0
        self.d_mu5 = 0.0
        self.d_mu6 = 0.0
        self.d_tau4 = 0.0
        self.d_tau5 = 0.0
        self.d_tau6 = 0.0
        self.d_44 = 0.0
        self.d_45 = 0.0
        self.d_46 = 0.0
        self.d_55 = 0.0
        self.d_56 = 0.0
        self.d_66 = 0.0

        self.ceV = 0.0
        self.ceA = 0.0
        self.cuV = 0.0
        self.cuA = 0.0
        self.cdV = 0.0
        self.cdA = 0.0

        self.deV = 0.0
        self.deA = 0.0
        self.duV = 0.0
        self.duA = 0.0
        self.ddV = 0.0
        self.ddA = 0.0
        
        self.deS = 0.0
        self.deP = 0.0
        self.duS = 0.0
        self.duP = 0.0
        self.ddS = 0.0
        self.ddP = 0.0

        self.cprotonV = None
        self.cneutronV = None
        self.cprotonA = None
        self.cneutronA = None
        self.dprotonV = None
        self.dneutronV = None
        self.dprotonA = None
        self.dneutronA = None
        self.dprotonS = None
        self.dneutronS = None
        self.dprotonP = None
        self.dneutronP = None

        # Now assign parameters based on user input
        self.__dict__.update(model_params)

        # initialize spectrum of HNLs
        self._initialize_spectrum()

        # and set vertices
        self.set_vertices()

    def set_vertices(self):

        ####################################################
        # SM Z boson couplings
        self.c_aj = np.array(
            [
                [const.gweak / 2 / const.cw, 0, 0, self.c_e4, self.c_e5, self.c_e6],
                [0, const.gweak / 2 / const.cw, 0, self.c_mu4, self.c_mu5, self.c_mu6],
                [0, 0, const.gweak / 2 / const.cw, self.c_tau4, self.c_tau5, self.c_tau6],
                [0, 0, 0, self.c_44, self.c_45, self.c_46],
                [0, 0, 0, self.c_45, self.c_55, self.c_56],
                [0, 0, 0, self.c_46, self.c_56, self.c_66],
            ]
        )

        ####################################################
        # Z' vector couplings
        self.d_aj = np.array(
            [
                [0, 0, 0, self.d_e4, self.d_e5, self.d_e6],
                [0, 0, 0, self.d_mu4, self.d_mu5, self.d_mu6],
                [0, 0, 0, self.d_tau4, self.d_tau5, self.d_tau6],
                [0, 0, 0, self.d_44, self.d_45, self.d_46],
                [0, 0, 0, self.d_45, self.d_55, self.d_56],
                [0, 0, 0, self.d_46, self.d_56, self.d_66],
            ]
        )

        ####################################################
        # h' scalar couplings
        self.s_aj = np.array(
            [
                [0, 0, 0, self.s_e4, self.s_e5, self.s_e6],
                [0, 0, 0, self.s_mu4, self.s_mu5, self.s_mu6],
                [0, 0, 0, self.s_tau4, self.s_tau5, self.s_tau6],
                [0, 0, 0, self.s_44, self.s_45, self.s_46],
                [0, 0, 0, self.s_45, self.s_55, self.s_56],
                [0, 0, 0, self.s_46, self.s_56, self.s_66],
            ]
        )

        ####################################################
        # create the transition mag moment scope
        self.t_aj = np.array(
            [
                [0, 0, 0, self.mu_tr_e4, self.mu_tr_e5, self.mu_tr_e6],
                [0, 0, 0, self.mu_tr_mu4, self.mu_tr_mu5, self.mu_tr_mu6],
                [0, 0, 0, self.mu_tr_tau4, self.mu_tr_tau5, self.mu_tr_tau6],
                [0, 0, 0, self.mu_tr_44, self.mu_tr_45, self.mu_tr_46],
                [0, 0, 0, self.mu_tr_45, self.mu_tr_55, self.mu_tr_56],
                [0, 0, 0, self.mu_tr_46, self.mu_tr_56, self.mu_tr_66],
            ]
        )

        prettyprinter.info(f"Model:{self._spectrum}")

        ####################################################
        # Nucleon couplings
        # n.b. lepton vertices already defined
        # for TMM, we already know it has to be (e*charge)

        if self.cprotonV is None:
            self.cprotonV = 2 * self.cuV + self.cdV
        if self.cprotonA is None:
            self.cprotonA = 2 * self.cuA + self.cdA
        if self.cneutronV is None:
            self.cneutronV = 2 * self.cdV + self.cuV
        if self.cneutronA is None:
            self.cneutronA = 2 * self.cdA + self.cuA

        if self.dprotonV is None:
            self.dprotonV = 2 * self.duV + self.ddV
        if self.dprotonA is None:
            self.dprotonA = 2 * self.duA + self.ddA
        if self.dneutronV is None:
            self.dneutronV = 2 * self.ddV + self.duV
        if self.dneutronA is None:
            self.dneutronA = 2 * self.ddA + self.duA

        if self.dprotonS is None:
            self.dprotonS = 2 * self.duS + self.ddS
        if self.dneutronS is None:
            self.dneutronS = 2 * self.ddS + self.duS
        if self.dprotonP is None:
            self.dprotonP = 2 * self.duP + self.ddP
        if self.dneutronP is None:
            self.dneutronP = 2 * self.ddP + self.duP

        self._update_spectrum()


class ThreePortalModel(HNLModel):
    def __init__(self, model_file=None, **model_params):

        super().__init__(model_file, model_params=model_params)

        # set defaults
        self.gD = 1.0
        self.alphaD = None

        self.epsilon = 1e-3
        self.epsilon2 = None
        self.chi = None
        self.alpha_epsilon2 = None

        self.epsilonZ = 0.0  # Z-Z' mass mixing

        self.theta = 0.0 # scalar mixing

        self.Ue4 = 0.0
        self.Ue5 = 0.0
        self.Ue6 = 0.0
        self.Umu4 = 1e-3
        self.Umu5 = 0.0
        self.Umu6 = 0.0
        self.Utau4 = 0.0
        self.Utau5 = 0.0
        self.Utau6 = 0.0
        self.UD4 = 1.0
        self.UD5 = 1.0
        self.UD6 = 1.0

        # Now assign parameters based on user input
        self.__dict__.update(model_params)

        # Now assign parameters based on user input
        self.__dict__.update(model_params)

        # initialize spectrum of HNLs
        self._initialize_spectrum()

        # and set vertices
        self.set_vertices()

    def set_vertices(self):
        """
        set all other variables starting from base members

        """

        # initialize spectrum
        self._initialize_spectrum()

        # create the scalar couplings
        self.s_aj = np.array(
            [
                [0, 0, 0, self.s_e4, self.s_e5, self.s_e6],
                [0, 0, 0, self.s_mu4, self.s_mu5, self.s_mu6],
                [0, 0, 0, self.s_tau4, self.s_tau5, self.s_tau6],
                [0, 0, 0, self.s_44, self.s_45, self.s_46],
                [0, 0, 0, self.s_45, self.s_55, self.s_56],
                [0, 0, 0, self.s_46, self.s_56, self.s_66],
            ]
        )

        # create the transition mag moment scope
        self.t_aj = np.array(
            [
                [0, 0, 0, self.mu_tr_e4, self.mu_tr_e5, self.mu_tr_e6],
                [0, 0, 0, self.mu_tr_mu4, self.mu_tr_mu5, self.mu_tr_mu6],
                [0, 0, 0, self.mu_tr_tau4, self.mu_tr_tau5, self.mu_tr_tau6],
                [0, 0, 0, self.mu_tr_44, self.mu_tr_45, self.mu_tr_46],
                [0, 0, 0, self.mu_tr_45, self.mu_tr_55, self.mu_tr_56],
                [0, 0, 0, self.mu_tr_46, self.mu_tr_56, self.mu_tr_66],
            ]
        )

        prettyprinter.info(f"Model:{self._spectrum}")

        # Assign the correct value of kinetic mixing and gD given user input
        if self.chi is not None:
            self.epsilon = self.epsilon * const.cw
        elif self.epsilon2 is not None:
            self.epsilon = np.sqrt(self.epsilon2)
            self.chi = self.epsilon / const.cw
        elif self.alpha_epsilon2 is not None:
            self.epsilon = np.sqrt(self.alpha_epsilon2 / const.alphaQED)
            self.chi = self.epsilon / const.cw
        elif self.epsilon is not None:
            self.chi = self.epsilon / const.cw

        # dark coupling
        if self.alphaD is not None:
            self.gD = np.sqrt(4 * np.pi * self.alphaD)
        elif self.gD is not None:
            pass

        # CHARGED FERMION VERTICES
        # all the following is true to leading order in chi
        self.tanchi = math.tan(self.chi)
        self.sinof2chi = 2 * self.tanchi / (1.0 + self.tanchi ** 2)
        self.cosof2chi = (1.0 - self.tanchi ** 2) / (1.0 + self.tanchi ** 2)
        self.s2chi = (1.0 - self.cosof2chi) / 2.0
        self.c2chi = 1 - self.s2chi

        entry_22 = self.c2chi - const.s2w * self.s2chi - (self.mzprime / const.m_Z) ** 2
        self.tanof2beta = const.sw * self.sinof2chi / (entry_22)
        self.beta = const.sw * self.chi
        self.sinof2beta = const.sw * self.sinof2chi / np.sqrt(entry_22 ** 2 + self.sinof2chi ** 2 * const.s2w)
        self.cosof2beta = entry_22 / np.sqrt(entry_22 ** 2 + self.sinof2chi ** 2 * const.s2w)

        # #####################
        if self.tanof2beta != 0:
            self.tbeta = self.sinof2beta / (1 + self.cosof2beta)
        else:
            self.tbeta = 0.0

        self.sbeta = math.sqrt((1 - self.cosof2beta) / 2.0) * np.sign(self.sinof2beta)  # FIX ME -- works only for |beta| < pi/2
        self.cbeta = math.sqrt((1 + self.cosof2beta) / 2.0)  # FIX ME -- works only for |beta| < pi/2

        # some abbreviations
        self._weak_vertex = const.gweak / const.cw / 2.0
        self._gschi = self.gD * self.sbeta
        # dark couplings acquired by Z boson
        self._g_weak_correction = self.cbeta + self.tanchi * const.sw * self.sbeta
        self._g_dark_correction = self.cbeta * self.tanchi * const.sw - self.sbeta

        # Charged leptons
        self.ceV = self._weak_vertex * (self.cbeta * (2 * const.s2w - 0.5) + 3.0 / 2.0 * self.sbeta * const.sw * self.tanchi)
        self.ceA = self._weak_vertex * (-(self.cbeta + self.sbeta * const.sw * self.tanchi) / 2.0)
        # self.ceV = weak_vertex*(const.gweak/(2*const.cw) * (2*const.s2w - 0.5))
        # self.ceA = weak_vertex*(const.gweak/(2*const.cw) * (-1.0/2.0))

        # quarks
        self.cuV = self._weak_vertex * (self.cbeta * (0.5 - 4 * const.s2w / 3.0) - 5.0 / 6.0 * self.sbeta * const.sw * self.tanchi)
        self.cuA = self._weak_vertex * ((self.cbeta + self.sbeta * const.sw * self.tanchi) / 2.0)

        self.cdV = self._weak_vertex * (self.cbeta * (-0.5 + 2 * const.s2w / 3.0) + 1.0 / 6.0 * self.sbeta * const.sw * self.tanchi)
        self.cdA = self._weak_vertex * (-(self.cbeta + self.sbeta * const.sw * self.tanchi) / 2.0)

        self.deV = self._weak_vertex * (3.0 / 2.0 * self.cbeta * const.sw * self.tanchi - self.sbeta * (-0.5 + 2 * const.s2w))
        self.deA = self._weak_vertex * ((self.sbeta - self.cbeta * const.sw * self.tanchi) / 2.0)

        self.duV = -self._weak_vertex * (-5.0 / 6.0 * self.cbeta * const.sw * self.tanchi - self.sbeta * (0.5 - 4.0 / 3.0 * const.s2w))
        self.duA = self._weak_vertex * ((-self.sbeta + self.cbeta * const.sw * self.tanchi) / 2.0)

        self.ddV = -self._weak_vertex * (-self.sbeta * (-0.5 + 2 / 3.0 * const.s2w) + 1.0 / 6.0 * self.cbeta * const.sw * self.tanchi)
        self.ddA = self._weak_vertex * ((self.sbeta - self.cbeta * const.sw * self.tanchi) / 2.0)

        self.dprotonV = 2 * self.duV + self.ddV
        self.dprotonA = 2 * self.duA + self.ddA
        self.dneutronV = 2 * self.ddV + self.duV
        self.dneutronA = 2 * self.ddA + self.duA

        self.cprotonV = 2 * self.cuV + self.cdV
        self.cprotonA = 2 * self.cuA + self.cdA
        self.cneutronV = 2 * self.cdV + self.cuV
        self.cneutronA = 2 * self.cdA + self.cuA

        # ###################################################
        # NEUTRAL FERMION VERTICES
        self.Ue = [1, 0, 0, self.Ue4, self.Ue5, self.Ue6]
        self.Umu = [0, 1, 0, self.Umu4, self.Umu5, self.Umu6]
        self.Utau = [0, 0, 1, self.Utau4, self.Utau5, self.Utau6]
        self.Udark = [0, 0, 0, self.UD4, self.UD5, self.UD6]

        # FIX-ME -- expand to arbitrary number of dark flavors?
        self.n_dark_HNLs = 1  
        
        # list of dark flavors
        self.inds_dark = range(const.ind_tau + 1, 3 + self.n_dark_HNLs)

        # Mixing matrices
        # if PMNS, use dtype=complex
        self.Ulep = np.diag(np.full_like(self.Ue, 1))
        # self.Ulep = np.diag(np.full(self.n_nus,1,dtype=complex))
        # self.Ulep[:3,:3] = const.UPMNS # massless light neutrinos

        # loop over HNL indices
        for i in range(3, self.n_HNLs + 3):
            self.Ulep[const.ind_e, i] = self.Ue[i]
            self.Ulep[i, const.ind_e] = self.Ue[i]

            self.Ulep[const.ind_mu, i] = self.Umu[i]
            self.Ulep[i, const.ind_mu] = self.Umu[i]

            self.Ulep[const.ind_tau, i] = self.Utau[i]
            self.Ulep[i, const.ind_tau] = self.Utau[i]

            self.Ulep[self.inds_dark, i] = self.Udark[i] / self.n_dark_HNLs
            self.Ulep[i, self.inds_dark] = self.Udark[i] / self.n_dark_HNLs

        self.Ua = self.Ulep[const.inds_active, :]
        self.Uactive_heavy = self.Ulep[const.inds_active, 3:]

        self.UD = self.Ulep[self.inds_dark, :]
        self.UD_heavy = self.Ulep[self.inds_dark, 3:]

        # Matrices
        # (Ua^dagger . Ua)_ij = Uei Uej + Umui Umuj + Utaui Utauj
        self.C_weak = np.dot(self.Ua.conjugate().T, self.Ua)
        # (UDi^dagger . UD)_ij
        self.D_dark = np.dot(self.UD.conjugate().T, self.UD)

        # Vectors
        # ( |Ua4|^2, |Ua5|^2, |Ua6|^2, ...)
        self.UactiveUactive_diag = np.diagonal(np.dot(self.Uactive_heavy.conjugate(), self.Uactive_heavy.T))
        # ( |UD4|^2, |UD5|^2, |UD6|^2, ...)
        self.UDUD_diag = np.diagonal(np.dot(self.UD_heavy.conjugate(), self.UD_heavy.T))
        # (Ua4* UD4, Ua5* UD5, Ua6* UD6,..)
        self.UactiveUD_diag = np.diagonal(np.dot(self.Uactive_heavy.conjugate(), self.UD_heavy.T))

        # ( |Ue4|^2 +  |Ue5|^2 + ... , |Umu4|^2 + |Umu5|^2 + ..., |Utau4|^2 + |Utau5|^2 + ...)
        self.UahUah_mass_summed = np.sum(np.dot(self.Uactive_heavy.conjugate(), self.Uactive_heavy.T), axis=0)
        # (Ue4* UD4 + Ue5* UD5+... , Umu4* Umu4 + Umu5* Umu5+..., Utau4* Utau4 + Utau5* Utau5+...)
        self.UDhUDh_mass_summed = np.sum(np.dot(self.UD_heavy.conjugate(), self.UD_heavy.T), axis=0)
        # (Ue4* UD4 + Ue5* UD5+... , Umu4* Umu4 + Umu5* Umu5+..., Utau4* Utau4 + Utau5* Utau5+...)
        self.UahUDh_mass_summed = np.sum(np.dot(self.Uactive_heavy.conjugate(), self.UD_heavy.T), axis=0)

        # Numbers
        # |U_a4|^2 + |U_{a5}|^2 + ...
        self.A_heavy_sum = np.sum(self.UactiveUactive_diag)
        # |U_D4|^2 + |U_{D5}|^2 + ...
        self.D_heavy_sum = np.sum(self.UDUD_diag)
        # U_A4* U_D4 + U_A5* U_D5 + ..
        self.AD_heavy_sum = np.sum(self.UactiveUD_diag)

        self.A4 = self.Ue4 ** 2 + self.Umu4 ** 2 + self.Utau4 ** 2
        self.A5 = self.Ue5 ** 2 + self.Umu5 ** 2 + self.Utau5 ** 2
        self.A6 = self.Ue6 ** 2 + self.Umu6 ** 2 + self.Utau6 ** 2

        self.D4 = self.UD4 ** 2  # self.UD4#(1.0 - self.A4 - self.A5)/(1.0+self.R)
        self.D5 = self.UD5 ** 2  # self.UD5#(1.0 - self.A4 - self.A5)/(1.0+1.0/self.R)
        self.D6 = self.UD6 ** 2  # self.UD6#(1.0 - self.A4 - self.A5)/(1.0+1.0/self.R)

        # NEUTRAL LEPTON SECTOR VERTICES
        self.c_ij = self._weak_vertex * self.C_weak * self._g_weak_correction + self.D_dark * self._gschi

        self.c_aj = (
            self._weak_vertex * np.dot(np.diag(1 - self.UahUah_mass_summed), self.Uactive_heavy) * self._g_weak_correction
            + np.dot(np.diag(-self.UahUDh_mass_summed), self.UD_heavy) * self._gschi
        )

        self.d_ij = self._weak_vertex * self.C_weak * self._g_dark_correction + self.D_dark * self.gD
        self.d_aj = (
            self._weak_vertex * np.dot(np.diag(1 - self.UahUah_mass_summed), self.Uactive_heavy) * self._g_dark_correction
            + np.dot(np.diag(-self.UahUDh_mass_summed), self.UD_heavy) * self.gD
        )

        # make it 3 x n_nus
        self.c_aj = np.hstack((np.diag([1, 1, 1]), self.c_aj))
        self.d_aj = np.hstack((np.diag([1, 1, 1]), self.d_aj))

        # make n_nus x n_nus
        self.c_aj = np.vstack((self.c_aj, self.c_ij[3:]))
        self.d_aj = np.vstack((self.d_aj, self.d_ij[3:]))
        self.dlight = 0.0

        # ########################
        # Scalar couplings
        self.tantheta = np.tan(self.theta)
        self.sintheta = np.sin(self.theta)
        self.costheta = np.cos(self.theta)

        # light quark couplings determine higgs coupling to nucleon
        # see e.g. arxiv.org/abs/1306.4710
        self.sigma_l = 0.058  # GeV
        self.sigma_0 = 0.055  # GeV
        z = 1.49  # isospin breaking parameter
        y = 1 - self.sigma_0 / self.sigma_l
        _prefactor = 1 / (const.m_u + const.m_d) * self.sigma_l / const.m_avg
        self.fu = _prefactor * const.m_u * (2 * z + y * (1 - z)) / (1 + z)
        self.fd = _prefactor * const.m_d * (2 - y * (1 - z)) / (1 + z)
        self.fs = _prefactor * const.m_s * y

        self.fN_higgs = 2 / 9 + 7 / 9 * (self.fu + self.fd + self.fs)
        self.c_nucleon_higgs = self.fN_higgs * const.m_avg / const.vev_EW
        self.cnucleonS = self.sintheta * self.c_nucleon_higgs
        
        # isospin
        self.dprotonS = self.cnucleonS
        self.dneutronS = self.cnucleonS
        self.ceS = self.costheta * const.m_e / const.vev_EW / np.sqrt(2)
        self.deS = self.sintheta * const.m_e / const.vev_EW / np.sqrt(2)

        # no pseudo-scalar coupling
        self.deP = 0.0

        self._update_spectrum()

# For complete HNL decay rates
"""
    def compute_rates(self):

            ## FIX ME -- GENERALIZE TO N5 and N6
            ##################
            # Neutrino 4
            mh = self.m4
            rates = {}
            neutrinos = [lp.nu_e, lp.nu_mu, lp.nu_tau]

            # channels with 3 neutrinos in final state
            rates['nu_nu_nu'] = 0.0
            for nu_a in neutrinos:
                rates['nu_nu_nu'] += dr.nui_nuj_nuk_nuk(self, const.N4, nu_a)

            # channels with 1 neutrino in final states
            rates['nu_gamma'] = 0
            rates['nu_e_e'] = 0
            rates['nu_mu_mu'] = 0
            rates['nu_e_mu'] = 0
            rates['nu_pi'] = 0
            rates['nu_eta'] = 0
            rates['e_pi'] = 0
            rates['e_K'] = 0
            rates['mu_pi'] = 0
            rates['mu_K'] = 0

            for nu_a in neutrinos:          # nu gamma
                rates['nu_gamma'] += dr.nui_nuj_gamma(self, const.N4, nu_a)
                # dileptons -- already contains the Delta L = 2 channel
                if mh > 2*lp.e_minus.mass/1e3:
                    rates['nu_e_e'] += dr.nui_nuj_ell1_ell2(self, const.N4, nu_a, lp.e_minus, lp.e_plus)
                if mh > lp.e_minus.mass/1e3 + lp.mu_minus.mass/1e3:
                    rates['nu_e_mu'] += dr.nui_nuj_ell1_ell2(self, const.N4, nu_a, lp.e_minus, lp.mu_plus)
                    rates['nu_e_mu'] += dr.nui_nuj_ell1_ell2(self, const.N4, nu_a, lp.mu_minus, lp.e_plus)
                if mh > 2*lp.mu_minus.mass/1e3:
                    rates['nu_mu_mu'] += dr.nui_nuj_ell1_ell2(self, const.N4, nu_a, lp.mu_minus, lp.mu_plus)
                # pseudoscalar -- neutral current
                if mh > lp.pi_0.mass/1e3:
                    rates['nu_pi'] += dr.nui_nu_P(self, const.N4, nu_a, lp.pi_0)
                if mh > lp.eta.mass/1e3:
                    rates['nu_eta'] += dr.nui_nu_P(self, const.N4, nu_a, lp.eta)

            # CC-only channels
            # pseudoscalar -- factor of 2 for delta L=2 channel
            if mh > lp.e_minus.mass/1e3+lp.pi_plus.mass/1e3:
                rates['e_pi'] = dr.nui_l_P(self, const.N4, lp.e_minus, lp.pi_plus)
            if mh > lp.e_minus.mass/1e3+lp.K_plus.mass/1e3:
                rates['e_K'] = dr.nui_l_P(self, const.N4, lp.e_minus, lp.K_plus)

            # pseudoscalar -- already contain the Delta L = 2 channel
            if mh > lp.mu_minus.mass/1e3+lp.pi_plus.mass/1e3:
                rates['mu_pi'] = dr.nui_l_P(self, const.N4, lp.mu_minus, lp.pi_plus)
            if mh > lp.mu_minus.mass/1e3+lp.K_plus.mass/1e3:
                rates['mu_K'] = dr.nui_l_P(self, const.N4, lp.mu_minus, lp.K_plus)

            self.rates = rates

            # total decay rate
            self.rate_total = sum(self.rates.values())

            # total decay rate
            self.lifetime = const.get_decay_rate_in_s(self.rate_total)
            self.ctau0 = const.get_decay_rate_in_cm(self.rate_total)

            # Branchin ratios
            brs = {}
            for channel in self.rates.keys():
                brs[channel] = self.rates[channel]/self.rate_total
            self.brs = brs


class HNLparticle():

    def __init__(self, this_hnl, bsm_model):

        self.this_hnl = this_hnl
        self.bsm_model = bsm_model

        # Dirac or Majorana
        self.HNLtype    = bsm_model.HNLtype

        self.mHNL = bsm_model.masses[get_HNL_index(this_hnl)]


    def _setup_rates(self):

        self.rates = {}
        self.daughter_neutrinos = self.bsm_model.nu_spectrum[:pdg.get_HNL_index(self.this_hnl)]


        # Setting all decay BRs to 0
        for nu_i in self.daughter_neutrinos:
            for nu_j in self.daughter_neutrinos:
                for nu_k in self.daughter_neutrinos:
                    self.rates[f'{nu_i.name}_{nu_j.name}_{nu_k.name}'] = 0
            self.rates[f'{nu_i.name}_{lp.gamma.name}']                     = 0
            self.rates[f'{nu_i.name}_{lp.e_plus.name}_{lp.e_minus.name}']  = 0
            self.rates[f'{nu_i.name}_{lp.mu_plus.name}_{lp.e_minus.name}'] = 0
            self.rates[f'{nu_i.name}_{lp.e_plus.name}_{lp.mu_minus.name}'] = 0
            self.rates[f'{nu_i.name}_{mu_plus.name}_{mu_minus.name}']      = 0
            self.rates[f'{nu_i.name}_{lp.pi_0.name}']                      = 0
            self.rates[f'{nu_i.name}_{lp.eta.name}']                       = 0
            self.rates[f'{lp.e_minus.name}_{lp.pi_plus.name}']             = 0
            self.rates[f'{lp.e_plus.name}_{lp.pi_minus.name}']             = 0
            self.rates[f'{lp.e_minus.name}_{lp.K_plus.name}']              = 0
            self.rates[f'{lp.e_plus.name}_{lp.K_minus.name}']              = 0
            self.rates[f'{lp.mu_minus.name}_{lp.pi_plus.name}']            = 0
            self.rates[f'{lp.mu_plus.name}_{lp.pi_minus.name}']            = 0
            self.rates[f'{lp.mu_minus.name}_{lp.K_plus.name}']             = 0
            self.rates[f'{lp.mu_plus.name}_{lp.K_minus.name}']             = 0


    def compute_rates(self):


        #################
        mh = self.mHNL

        # channels with 3 neutrinos in final state
        for nu_i in self.daughter_neutrinos:
            for nu_j in self.daughter_neutrinos:
                for nu_k in self.daughter_neutrinos:
                    self.rates[f'{nu_i.name}_{nu_j.name}_{nu_k.name}'] += dr.new_nuh_nui_nuj_nuk(
                                                                    self.bsm_model,
                                                                    initial_neutrino=self.particle,
                                                                    final_neutrinoi=nu_a,
                                                                    final_neutrinoj=nu_b,
                                                                    final_neutrinok=nu_c)
        for nu_i in self.daughter_neutrinos:
            # nu gamma
            self.rates[f'{nu_i.name}_{lp.gamma.name}'] += dr.nui_nuj_gamma(self.bsm_model, self.this_hnl, nu_i)
            # dileptons -- already contains the Delta L = 2 channel
            # e+e-
            if mh > 2*lp.e_minus.mass/1e3:
                self.rates[f'{nu_i.name}_{lp.e_plus.name}_{lp.e_minus.name}'] += dr.nui_nuj_ell1_ell2(self.bsm_model, self.this_hnl, nu_i, lp.e_minus, lp.e_plus)
            # e+ mu- and # e- mu+
            if mh > lp.e_minus.mass/1e3 + lp.mu_minus.mass/1e3:
                self.rates[f'{nu_i.name}_{lp.mu_plus.name}_{lp.e_minus.name}'] += dr.nui_nuj_ell1_ell2(self.bsm_model, self.this_hnl, nu_i, lp.e_minus, lp.mu_plus)
                self.rates[f'{nu_i.name}_{lp.e_plus.name}_{lp.mu_minus.name}'] += dr.nui_nuj_ell1_ell2(self.bsm_model, self.this_hnl, nu_i, lp.mu_minus, lp.e_plus)
            # mu+ mu-
            if mh > 2*lp.mu_minus.mass/1e3:
                self.rates[f'{nu_i.name}_{mu_plus.name}_{mu_minus.name}'] += dr.nui_nuj_ell1_ell2(self.bsm_model, self.this_hnl, nu_i, lp.mu_minus, lp.mu_plus)
            # pseudoscalar -- neutral current
            if mh > lp.pi_0.mass/1e3:
                self.rates[f'{nu_i.name}_{lp.pi_0.name}'] += dr.nui_nu_P(self.bsm_model, self.this_hnl, nu_i, lp.pi_0)
            if mh > lp.eta.mass/1e3:
                self.rates[f'{nu_i.name}_{lp.eta.name}'] += dr.nui_nu_P(self.bsm_model, self.this_hnl, nu_i, lp.eta)


        # CC-only channels
        # pseudoscalar
        if mh > lp.e_minus.mass/1e3+lp.pi_plus.mass/1e3:
            self.rates[f'{lp.e_minus.name}_{lp.pi_plus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.e_minus, lp.pi_plus)
            self.rates[f'{lp.e_plus.name}_{lp.pi_minus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.e_plus, lp.pi_minus)
        if mh > lp.e_minus.mass/1e3+lp.K_plus.mass/1e3:
            self.rates[f'{lp.e_minus.name}_{lp.K_plus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.e_minus, lp.K_plus)
            self.rates[f'{lp.e_plus.name}_{lp.K_minus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.e_plus, lp.K_minus)

        # pseudoscalar
        if mh > lp.mu_minus.mass/1e3+lp.pi_plus.mass/1e3:
            self.rates[f'{lp.mu_minus.name}_{lp.pi_plus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.mu_minus, lp.pi_plus)
            self.rates[f'{lp.mu_plus.name}_{lp.pi_minus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.mu_plus, lp.pi_minus)
        if mh > lp.mu_minus.mass/1e3+lp.K_plus.mass/1e3:
            self.rates[f'{lp.mu_minus.name}_{lp.K_plus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.mu_minus, lp.K_plus)
            self.rates[f'{lp.mu_plus.name}_{lp.K_minus.name}'] = nui_l_P(self.bsm_model, self.this_hnl, lp.mu_plus, lp.K_minus)


        # total decay rate
        self.rate_total = sum(self.rates.values())

        # total decay rate
        self.lifetime = const.get_decay_rate_in_s(self.rate_total)
        self.ctau0 = const.get_decay_rate_in_cm(self.rate_total)

        # branching ratios
        brs = {}
        for channel in self.rates.keys():
            brs[channel] = self.rates[channel]/self.rate_total
        self.brs = brs

"""


#!/usr/bin/env python3

# proceed as usual
from setuptools import setup, find_packages, Extension
import numpy as np
import os

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


extensions = [
    Extension("DarkNews.Cfourvec", ["src/DarkNews/Cfourvec.pyx"]),
]
CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 0))) and cythonize is not None

if CYTHONIZE:
    compiler_directives = {"language_level": 3, "embedsignature": True}
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)


setup_args = dict(
    # ext_modules = cythonize(["src/DarkNews/Cfourvec.pyx"]),
    ext_modules=extensions,
    include_dirs=np.get_include(),
)


if __name__ == "__main__":
    setup(**setup_args)
